"""
202012

General purpose plotting from arbitrary csv files

Started as specifically for SanPy (using reanalyze.py) from a database of spikes

20210112, extending it to any csv for use with bImPy nodes/edges
    See: bImPy/bimpy/analysis/bScrapeNodesAndEdges.py

I want this to run independent of SanPy, to do this:
    copy (pandas model, checkbox delegate) from sanpy.butil
    into a local folder and change imports

Requires (need to make local copies of (pandas model, checkbox delegate)
    pandas
    numpy
    PyQt5
    matplotlib
    seaborn
    mplcursors
    openpyxl # to load xlsx
"""

import os, sys, io, csv
from collections import OrderedDict
import traceback
import pandas as pd
import numpy as np
from PyQt5 import QtCore, QtGui, QtWidgets
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT
from matplotlib.figure import Figure
import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt # abb 202012 added to set theme
import seaborn as sns
import mplcursors # popup on hover

#sysPath = os.path.dirname(sys.path[0])
#sysPath = os.path.join(sysPath, 'sanpy')
#print('bScatterPlotWidget2.py sysPath:', sysPath)
#sys.path.append(sysPath)

#print('os.getcwd():', os.getcwd())
#import sanpy
#import bUtil # from sanpy folder, for pandas model

def old_loadDatabase(path):
    """
    path: full path to .csv file generated with reanalyze.py
    """
    #path = '/Users/cudmore/data/laura-ephys/Superior vs Inferior database_master.csv'
    masterDf = None
    if not os.path.isfile(path):
        print('eror: bUtil.loadDatabase() did not find file:', path)
    elif path.endswith('.csv'):
        masterDf = pd.read_csv(path, header=0) #, dtype={'ABF File': str})
    elif path.endswith('.xls'):
        masterDf = pd.read_excel(path, header=0) #, dtype={'ABF File': str})
    elif path.endswith('.xlsx'):
        masterDf = pd.read_excel(path, header=0, engine='openpyxl') #, dtype={'ABF File': str})
    else:
        print('error: file type not supported. Expecting csv/xls/xlsx. Path:', path)

    #self.masterDfColumns = self.masterDf.columns.to_list()

    # not sure what this was for ???
    # 20210112, put back in if necc
    #self.masterCatColumns = ['Condition', 'File Number', 'Sex', 'Region', 'filename', 'analysisname']
    #self.masterCatColumns = self.categoricalList

    #print(self.masterDf.head())
    '''
    print(masterDf.info())
    print('masterDf.iloc[0,3]:', masterDf.iloc[0,3], type(masterDf.iloc[0,3]))
    print('start seconds:', masterDf['Start Seconds'].dtype.type)
    print('start seconds:', masterDf['Start Seconds'].dtype)
    '''
    #
    return masterDf

def printDict(d, withType=False):
    for k,v in d.items():
        if withType:
            print(f'  {k}: {v} {type(v)}')
        else:
            print(f'  {k}: {v}')

class myPandasModel(QtCore.QAbstractTableModel):

    def __init__(self, data):
        """
        data: pandas dataframe
        """
        QtCore.QAbstractTableModel.__init__(self)
        self.verbose = False
        self._data = data
        columnList = self._data.columns.values.tolist()
        if 'include' in columnList:
            self.includeCol = columnList.index('include')
        else:
            self.includeCol = None
        #print('pandasModel.__init__() self.includeCol:', self.includeCol)
        self.columns_boolean = ['include']

    def rowCount(self, parent=None):
        #if self.verbose: print('myPandasModel.rowCount()')
        return self._data.shape[0]

    def columnCount(self, parnet=None):
        #if self.verbose: print('myPandasModel.columnCount()')
        return self._data.shape[1]

    def data(self, index, role=QtCore.Qt.DisplayRole):
        if self.verbose: print('myPandasModel.data()')
        if index.isValid():
            if role == QtCore.Qt.DisplayRole:

                #return QtCore.QVariant()

                #return str(self._data.iloc[index.row(), index.column()])
                retVal = self._data.iloc[index.row(), index.column()]
                if isinstance(retVal, np.float64):
                    retVal = float(retVal)
                    retVal = round(retVal,4) # round everything to 4 decimal places
                    if np.isnan(retVal):
                        retVal = ''
                elif isinstance(retVal, np.int64):
                    retVal = int(retVal)
                #
                return retVal
            elif role == QtCore.Qt.BackgroundRole:
                #return
                return QtCore.QVariant()

        return None

    def update(self, dataIn):
        if self.verbose: print('myPandasModel.update()')

    def setData(self, index, value, role=QtCore.Qt.DisplayRole):
        """
        This is curently limited to only handle checkbox
        todo: extend to allow editing

        Returns:
            True if value is changed. Calls layoutChanged after update.
            False if value is not different from original value.
        """
        if self.verbose: print('myPandasModel.setData()')
        print('  myPandasModel.setData() row:', index.row(), 'column:', index.column(), 'value:', value, type(value))
        #if index.column() == self.includeCol:

        # dataChanged is inherited from QAbstractItemModel
        #topLeftIndex = index
        #bottomRightIndex = index
        #self.dataChanged.emit(index, index)

        if 1:

            #print('value:', value, type(value))
            v = self._data.iloc[index.row(), index.column()]
            #print('before v:',v, type(v))
            #print('isinstance:', isinstance(v, np.float64))
            if isinstance(v, np.float64):
                try:
                    value = float(value)
                except (ValueError) as e:
                    print('please enter a number')
                    return False

            # set
            self._data.iloc[index.row(), index.column()] = value

            v = self._data.iloc[index.row(), index.column()]
            print('    after v:',v, type(v))
            return True
        return True

    def flags(self, index):
        if self.verbose:
            print('myPandasModel.flags()')
            print('  index.column():', index.column())
        if 1:
            # turn on editing (limited to checkbox for now)
            if index.column() in self.columns_boolean:
                #print('  return with columns_boolean')
                return QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsUserCheckable
            #print('  return with ...')
            return QtCore.Qt.ItemIsEditable | QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable
        else:
            return QtCore.Qt.ItemIsEnabled

    def headerData(self, col, orientation, role):
        #if self.verbose: print('myPandasModel.headerData()')
        if orientation == QtCore.Qt.Horizontal and role == QtCore.Qt.DisplayRole:
            return self._data.columns[col]
        return None

# see: https://stackoverflow.com/questions/17748546/pyqt-column-of-checkboxes-in-a-qtableview
class myCheckBoxDelegate(QtWidgets.QItemDelegate):
    """
    A delegate that places a fully functioning QCheckBox cell of the column to which it's applied.
    """
    def __init__(self, parent):
        QtWidgets.QItemDelegate.__init__(self, parent)
        self.verbose = False
        if self.verbose: print('myCheckBoxDelegate.__init__()')

    def createEditor(self, parent, option, index):
        """
        Important, otherwise an editor is created if the user clicks in this cell.
        """
        if self.verbose: print('myCheckBoxDelegate.createEditor()')
        return None

    def paint(self, painter, option, index):
        """
        Paint a checkbox without the label.

        option: PyQt5.QtWidgets.QStyleOptionViewItem
        index: PyQt5.QtCore.QModelIndex
        """
        if self.verbose: print('myCheckBoxDelegate.paint()')
        #print('  option:', option, 'index:', index)
        #print('  index.data():', type(index.data()), index.data())
        #HasCheckIndicator = QtWidget.QStyleOptionViewItem.HasCheckIndicator
        # options.HasCheckIndicator returns hex 4, value of enum
        # how do i query it?
        #print('  ', option.ViewItemFeatures().HasCheckIndicator) # returns PyQt5.QtWidgets.QStyleOptionViewItem.ViewItemFeature
        #print('  ', option.features)
        #print('  ', index.data(QtCore.Qt.CheckStateRole)  )
        #state = index.data(QtCore.Qt.CheckStateRole)
        #print('  state:', state, 'option.HasCheckIndicator:', option.HasCheckIndicator)
        self.drawCheck(painter, option, option.rect, QtCore.Qt.Unchecked if int(index.data()) == 0 else QtCore.Qt.Checked)

    def editorEvent(self, event, model, option, index):
        '''
        Change the data in the model and the state of the checkbox
        if the user presses the left mousebutton and this cell is editable. Otherwise do nothing.
        '''
        if self.verbose: print('myCheckBoxDelegate.editorEvent()')
        if not int(index.flags() & QtCore.Qt.ItemIsEditable) > 0:
            return False

        if event.type() == QtCore.QEvent.MouseButtonRelease and event.button() == QtCore.Qt.LeftButton:
            # Change the checkbox-state
            self.setModelData(None, model, index)
            return True

        return False


    def setModelData(self, editor, model, index):
        '''
        The user wanted to change the old state in the opposite.
        '''
        if self.verbose:
            print('myCheckBoxDelegate.setModelData()')
            print('  editor:', editor)
            print('  model:', model)
            print('  index:', index)
            print('  index.data():', type(index.data()), index.data())
        #data = index.data()
        #if isinstance(data, str):
        #    return
        newValue = 1 if int(index.data()) == 0 else 0
        model.setData(index, newValue, QtCore.Qt.EditRole)

#class MainWindow(QtWidgets.QWidget):
myInterfaceDefaults = OrderedDict({
    'X Statistic': None,
    'Y Statistic': None,
    'Hue': None,
    'Group By': None,
})

class myTableView(QtWidgets.QTableView):
    def __init__(self, dataType, parent=None):
        """
        dataType: in ['All Spikes', 'File Mean']
        """
        super(myTableView, self).__init__(parent)

        self.dataType = dataType

        self.doIncludeCheckbox = False # todo: turn this on
        self.keepCheckBoxDelegate = myCheckBoxDelegate(None)

        #self.setFont(QtGui.QFont('Arial', 10))
        self.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                  QtWidgets.QSizePolicy.Expanding)
        self.setSelectionBehavior(QtWidgets.QTableView.SelectRows)

        p = self.palette();
        color1 = QtGui.QColor('#dddddd')
        color2 = QtGui.QColor('#ffffff')
        p.setColor(QtGui.QPalette.Base, color1);
        p.setColor(QtGui.QPalette.AlternateBase, color2);
        self.setAlternatingRowColors(True)
        self.setPalette(p);

    def slotSelectRow(self, selectDict):
        """
        only selecct if selectDict['dataType'] matches what we are showing
        """
        if self.dataType != selectDict['dataType']:
            return

        print('slotSelectRow() selectDict:', selectDict)
        ind = selectDict['index']
        '''
        plotDf = selectDict['plotDf']
        index = plotDf.at[ind, 'index']
        index = int(index)
        '''
        index = ind
        index -= 1 # !!! MY VISUAL INDEX IN TABLE IS ONE BASED !!!
        column = 0
        modelIndex = self.model().index(index, column)
        self.setCurrentIndex(modelIndex)

    def slotSwitchTableDf(self, newDf):
        """
        set model from df
        """
        if newDf is None:
            model = None
        else:
            model = myPandasModel(newDf)
        self.slotSwitchTableModel(model)

    def slotSwitchTableModel(self, newModel):
        """
        switch between full .csv model and getMeanDf model
        """
        #print('myTableView.slotSwitchTableModel()')
        self.setModel(newModel)

        if newModel is None:
            return

        colList = newModel._data.columns.values.tolist()
        #print('  todo: slotSwitchTableModel() set all columns to default delegate itemDelegate()')
        for idx, col in enumerate(colList):
            # default delegate is self.tableView.itemDelegate()
            self.setItemDelegateForColumn(idx, self.itemDelegate())

        # install checkboxes in 'incude' column
        if self.doIncludeCheckbox and 'include' in colList:
            includeColIndex = colList.index('include')
            print(f'_switchTableModel() setting include column {includeColIndex} to myCheckBoxDelegate')
            #self.tableView.setItemDelegateForColumn(includeColIndex, myCheckBoxDelegate(None))
            self.setItemDelegateForColumn(includeColIndex, self.keepCheckBoxDelegate)

#class myMplCanvas(FigureCanvas):
class myMplCanvas(QtWidgets.QFrame):
    """
    hold an fig/plot canvas, in scatter plot we can have 1-4 of these
    """
    signalSelectFromPlot = QtCore.Signal(object)
    signalSelectSquare = QtCore.Signal(object, object)

    def __init__(self, plotNumber=None, parent=None):

        super(myMplCanvas, self).__init__(parent) # FigureCanvas

        #self.plotType = None

        #self.setFrameWidth(5)
        #self.setStyleSheet("background-color: rgb(0, 255, 0)")
        #self.setContentsMargins(0, 0, 0, 0)
        #self.setContentsMargins(-20, 0, 5, 0)
        self.setFrameShape(QtWidgets.QFrame.Box)
        self.setLineWidth(0)

        pal = self.palette()
        pal.setColor(QtGui.QPalette.WindowText, QtGui.QColor('red'))
        self.setPalette(pal)

        #
        self.plotNumber = plotNumber
        self.stateDict = None
        self.plotDf = None
        self.whatWeArePlotting = None
        self.mplCursorHover = None

        # needed to show canvas in widget
        self.layout = QtWidgets.QVBoxLayout() # any will do

        self.fig = Figure(constrained_layout=True)

        self.canvas = FigureCanvas(self.fig)

        self.canvas.axes = self.fig.add_subplot(111) # was this
        #self.canvas.axes = self.fig.add_axes([0.1, 0.1, 0.9, 0.9]) # [x, y, w, h]
        #gs1 = gridspec.GridSpec(1, 1)
        #gs1.tight_layout(self.fig, rect=[0.5, 0, 1, 1], h_pad=0.5)
        #self.canvas.axes = self.fig.add_subplot(gs1[0])

        self.cid2 = self.canvas.mpl_connect('pick_event', self.on_pick_event)

        self.scatterPlotSelection = None

        self.mplToolbar = NavigationToolbar2QT(self.canvas, self.canvas) # params are (canvas, parent)
        self.mplToolbar.hide() # initially hidden

        self.canvas.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                  QtWidgets.QSizePolicy.Expanding)
        self.canvas.updateGeometry()

        #fig.subplots_adjust(wspace=0.3, hspace=0.3)
        #self.fig.tight_layout(pad = 3.0) # Padding between the figure edge and the edges of subplots, as a fraction of the font size.
        #self.fig.subplots_adjust(left=0.15,right=0.9,
        #                            bottom=0.1,top=0.9,
        #                            hspace=0.2,wspace=0.2)

        #
        self.layout.addWidget(self.canvas)
        self.setLayout(self.layout)

    def buildUI(self):
        pass

    def mousePressEvent(self, event):
        print('myMplCanvas.mousePressEvent()')
        print('  todo: set all controls to match the plot just clicked on using self.stateDict!!!')
        #print('  stateDict:', self.stateDict)
        self.signalSelectSquare.emit(self.plotNumber, self.stateDict)

    def contextMenuEvent(self, event):
        print('myMplCanvas.contextMenuEvent()')
        contextMenu = QtWidgets.QMenu(self)
        saveAsAction = contextMenu.addAction("Save As...")
        action = contextMenu.exec_(self.mapToGlobal(event.pos()))
        if action == saveAsAction:
            print('todo: save as')
            self.fig.savefig('', dpi=600)

    def on_pick_event(self, event):
        try:
            print('myMplCanvas.on_pick_event() event:', event)
            print('  event.ind:', event.ind)

            if len(event.ind) < 1:
                return
            spikeNumber = event.ind[0]
            print('  selected:', spikeNumber)

            # propagate a signal to parent
            #self.myMainWindow.mySignal('select spike', data=spikeNumber)
            #self.selectSpike(spikeNumber)
        except (AttributeError) as e:
            pass

    def onPick(self, event):
        """
        when user clicks on a point in the graph

        todo: this makes perfect sense for scatter but maybe not other plots???
        """
        print('=== myMplCanvas.onPick()') #' event:', type(event), event)
        line = event.artist

        # filter out clicks on 'Annotation' used by mplcursors
        try:
            # when Scatter, line is 'PathCollection', a list of (x,y)
            offsets = line.get_offsets()
        except (AttributeError) as e:
            return

        ind = event.ind # ind is a list []
        if len(ind)==0:
            return
        ind = ind[0]

        # ind is the ith element in (x,y) list of offsets
        # ind 10 (0 based) is index 11 (1 based) in table list
        print(f'  selected from plot ind:{ind}, offsets values are {offsets[ind]}')
        selectDict = self.getAnnotation(ind)

        # to do, just put copy of state dict ???
        selectDict['plotType'] = self.stateDict['plotType']
        selectDict['dataType'] = self.stateDict['dataType']
        #
        # emit
        print('  myMplCanvas.signalSelectFromPlot.emit()', selectDict)
        self.signalSelectFromPlot.emit(selectDict)

    def _selectInd(self, ind):
        """
        visually select a point in scatter plot
        """
        print('myMplCanvas._selectInd() ind:', ind)
        if ind > len(self.plotDf)-1:
            return
        xVal = self.plotDf.at[ind, self.stateDict['xStat']]
        yVal = self.plotDf.at[ind, self.stateDict['yStat']]
        if self.scatterPlotSelection is not None:
            print('  scatterPlotSelection x:', xVal, 'y:', yVal)
            self.scatterPlotSelection.set_data(xVal, yVal)
        self.fig.canvas.draw()

    def getAnnotation(self, ind):
        if not np.issubdtype(ind, np.integer):
            print('myMplCanvas.getAnnotation() got bad ind:', ind, type(ind))
            return

        xStat = self.stateDict['xStat']
        yStat = self.stateDict['yStat']
        groupByColumnName = self.stateDict['groupByColumnName']

        analysisName = self.plotDf.at[ind, groupByColumnName]
        index = self.plotDf.at[ind, 'index']
        try:
            region = self.plotDf.at[ind, 'Region'] # not all will have this
        except (KeyError) as e:
            region = 'n/a'
        xVal = self.plotDf.at[ind, xStat]
        yVal = self.plotDf.at[ind, yStat]

        returnDict = {
                    'ind': ind,
                    'index': index,
                    'analysisName': analysisName,
                    'region': region,
                    'xVal': xVal,
                    'yVal': yVal,
                    #'plotDf': self.plotDf, # potentially very big
            }
        return returnDict

    def slotSelectSquare(self, plotNumber, stateDict):
        #print('myMplCanvas.slotSelectSquare()', plotNumber, 'self.plotNumber:', self.plotNumber)
        if plotNumber == self.plotNumber:
            self.setLineWidth(1)
        else:
            self.setLineWidth(0)

    def slotSelectInd(self, selectDict):
        if self.stateDict['plotType'] == selectDict['plotType']:
            # only select if same plot type, o/w selections are out of synch
            self._selectInd(selectDict['ind'])

    def slotCancelSelection(self):
        if self.scatterPlotSelection is not None:
            self.scatterPlotSelection.set_data([], [])
        #self.draw()
        # cancel mplCursorHover hover selection
        if self.mplCursorHover is not None:
            selections = self.mplCursorHover.selections
            if len(selections) ==1 :
                self.mplCursorHover.remove_selection(selections[0])

        # never cancel red square, we always wan't one
        #self.slotSelectSquare(None)

        #
        #self.draw() # to update hover
        self.fig.canvas.draw()

    def updateTheme(self):
        """
        redraw using stored self.state
        """
        self.myUpdate(stateDict=None)

    def myUpdateGlobal(self, stateDict):
        """
        update globals but do not plot
        """
        self.canvas.axes.legend().set_visible(stateDict['showLegend'])

        if stateDict['showMplToolbar']:
            self.mplToolbar.show()
        else:
            self.mplToolbar.hide()

        if stateDict['doHover'] and self.whatWeArePlotting is not None:
            self.mplCursorHover = mplcursors.cursor(self.whatWeArePlotting, hover=True)
            @self.mplCursorHover.connect("add")
            def _(sel):
                #sel.annotation.get_bbox_patch().set(fc="white")
                sel.annotation.arrow_patch.set(arrowstyle="simple", fc="white", alpha=.5)
                # row in df is from sel.target.index
                #print('sel.target.index:', sel.target.index)
                ind = sel.target.index
                annotationDict = self.getAnnotation(ind)
                myText = ''
                for k,v in annotationDict.items():
                    myText += f'{k}: {v}\n'
                sel.annotation.set_text(myText)
        elif not stateDict['doHover']:
            # cancel mplCursorHover hover selection
            if self.mplCursorHover is not None:
                selections = self.mplCursorHover.selections
                if len(selections) ==1 :
                    self.mplCursorHover.remove_selection(selections[0])
        #self.mplCursorHover = None

        #
        #self.draw() # to update hover
        self.fig.canvas.draw()

    def myUpdate(self, stateDict=None):
        """
        update plot based on control interface
        """

        # store stateDict so we can replot on changing dark theme
        if stateDict is None and self.stateDict is not None:
            # re-use our stateDict
            stateDict = self.stateDict
        else:
            self.stateDict = stateDict.copy()

        dataType = stateDict['dataType']
        hue = stateDict['hue']
        groupByColumnName = stateDict['groupByColumnName']

        plotType = stateDict['plotType']
        #self.plotType = plotType

        xStatHuman = stateDict['xStatHuman']
        yStatHuman = stateDict['yStatHuman']

        xStat = stateDict['xStat']
        yStat = stateDict['yStat']

        '''
        print('=== myMplCanvas.myUpdate()')
        print('  ', plotType)
        print('  ', 'xStatHuman:', xStatHuman, 'yStatHuman:', yStatHuman)
        print('  ', 'xStat:', xStat, 'yStat:', yStat)
        '''

        xIsCategorical = stateDict['xIsCategorical']
        yIsCategorical = stateDict['yIsCategorical']

        masterDf = stateDict['masterDf']
        meanDf = stateDict['meanDf']

        self.plotDf = meanDf

        self.canvas.axes.clear()

        picker = 5
        if plotType in ['Scatter Plot', 'Scatter + Raw + Mean']:
            # scatter plot user selection
            self.scatterPlotSelection, = self.canvas.axes.plot([], [], 'oy',
                                    markersize=12, fillstyle='none')

            # main scatter
            try:
                self.whatWeArePlotting = sns.scatterplot(x=xStat, y=yStat, hue=hue,
                                data=meanDf, ax=self.canvas.axes, picker=picker,
                                zorder=0)
            except (ValueError) as e:
                self.fig.canvas.draw()
                print('  EXCEPTION: in myUpdate() "Scatter Plot", exception is:')
                print('  ', e)
                print('  ', 'hue:', hue)

            # sem in both x and y, pulling from masterDf
            if dataType=='File Mean' or plotType=='Scatter + Raw + Mean':
                # we need to do this for each hue???
                # if x or y is in categorical (e.g. a string) then do not do this ...
                if xIsCategorical or yIsCategorical:
                    pass
                else:
                    print('  grabbing mean +- sem for self.groupByColumnName:', groupByColumnName)
                    color = 'k'
                    xd = masterDf.groupby(groupByColumnName).mean()[xStat]
                    xerrd = masterDf.groupby(groupByColumnName).sem()[xStat]
                    yd = masterDf.groupby(groupByColumnName).mean()[yStat]
                    yerrd = masterDf.groupby(groupByColumnName).sem()[yStat]
                    self.canvas.axes.errorbar(xd, yd, xerr=xerrd, yerr=yerrd,
                        fmt='none', capsize=0, zorder=10, color=color, alpha=0.5);

        elif plotType == 'Histogram':
            yStatHuman = 'Count'
            doKde = False #stateDict['doKDE']
            try:
                g = sns.histplot(x=xStat, hue=hue, kde=doKde,
                                data=meanDf, ax=self.canvas.axes, picker=picker)
            except (ValueError) as e:
                self.fig.canvas.draw()
                print('EXCEPTIONin Histogram:', e)

        elif plotType == 'Cumulative Histogram':
            yStatHuman = 'Probability'
            try:
                g = sns.histplot(x=xStat, hue=hue, cumulative=True, stat='density',
                                element="step", fill=False, common_norm=False,
                                data=meanDf, ax=self.canvas.axes, picker=picker)
            except (ValueError) as e:
                self.fig.canvas.draw()
                print('EXCEPTION in Cumulative Histogram:', e)

        elif plotType == 'Cumulative Histogram':
            yStatHuman = 'Probability'
            try:
                g = sns.histplot(x=xStat, hue=hue, cumulative=True, stat='density',
                                element="step", fill=False, common_norm=False,
                                data=meanDf, ax=self.canvas.axes, picker=picker)
            except (ValueError) as e:
                self.fig.canvas.draw()
                print('EXCEPTION in Cumulative Histogram:', e)

        elif plotType == 'Violin Plot':
            if not xIsCategorical:
                warningStr = 'Violin plot requires a categorical x statistic'
            else:
                g = sns.violinplot(x=xStat, y=yStat, hue=hue,
                        data=meanDf, ax=self.canvas.axes)

        elif plotType == 'Box Plot':
            if not xIsCategorical:
                warningStr = 'Box plot requires a categorical x statistic'
            else:
                g = sns.boxplot(x=xStat, y=yStat, hue=hue,
                        data=meanDf, ax=self.canvas.axes)

        elif plotType == 'Raw + Mean Plot':
            if not xIsCategorical:
                warningStr = 'Raw + Mean plot requires a categorical x statistic'
            else:
                try:
                    # does not work here for categorical x
                    #self.scatterPlotSelection, = self.canvas.axes[0].plot([], [], 'oy',
                    #                markersize=12, fillstyle='none')

                    '''
                    colorList = [('red'), ('green'), 'b', 'c', 'm', 'y']
                    hueList = meanDf[hue].unique()
                    palette = {}
                    for idx, hue in enumerate(hueList):
                        palette[hue] = colorList[idx]
                    print(palette)
                    '''

                    palette = sns.color_palette("Paired")
                    #palette = ['r', 'g', 'b']

                    # stripplot
                    #g = sns.swarmplot(x=xStat, y=yStat,
                    g = sns.stripplot(x=xStat, y=yStat,
                            hue=hue,
                            palette=palette,
                            data=meanDf,
                            ax=self.canvas.axes,
                            #color = color,
                            dodge=True,
                            alpha=0.6,
                            picker=picker,
                            zorder=1)


                    #g.legend().remove()
                    self.canvas.axes.legend().remove()

                    handles, labels = self.canvas.axes.get_legend_handles_labels()
                    l = self.canvas.axes.legend(handles[0:2], labels[0:2], bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)

                    '''
                    if self.darkTheme:
                        color = 'w'
                    else:
                        color = 'k'
                    color = [color] * len(hueList)
                    print('color:', color)
                    '''

                    self.whatWeArePlotting = sns.pointplot(x=xStat, y=yStat,
                            hue=hue,
                            #palette=palette,
                            data=meanDf,
                            estimator=np.nanmean,
                            ci=68, capsize=0.1,
                            ax=self.canvas.axes,
                            color='r',
                            #legend='full',
                            zorder=10)

                except (ValueError) as e:
                    print('EXCEPTION in "Raw + Mean Plot":', e)
                    traceback.print_exc()

        elif plotType == 'Regression Plot':
            # regplot does not have hue
            if xIsCategorical or yIsCategorical:
                warningStr = 'Regression plot requires continuous x and y statistics'
            else:
                # todo: loop and make a regplot
                # for each unique() name in
                # hue (like Region, Sex, Condition)
                hueList = masterDf[hue].unique()
                for oneHue in hueList:
                    if oneHue == 'None':
                        continue
                    tmpDf = meanDf [ meanDf[hue]==oneHue ]
                    #print('regplot oneHue:', oneHue, 'len(tmpDf)', len(tmpDf))
                    sns.regplot(x=xStat, y=yStat, data=tmpDf,
                            ax=self.canvas.axes);
        else:
            print('  did not understand plot type:', plotType)


        #
        # update
        self.canvas.axes.figure.canvas.mpl_connect("pick_event", self.onPick)

        self.mplCursorHover = None
        if stateDict['doHover'] and self.whatWeArePlotting is not None:
            self.mplCursorHover = mplcursors.cursor(self.whatWeArePlotting, hover=True)
            @self.mplCursorHover.connect("add")
            def _(sel):
                #sel.annotation.get_bbox_patch().set(fc="white")
                sel.annotation.arrow_patch.set(arrowstyle="simple", fc="white", alpha=.5)
                # row in df is from sel.target.index
                #print('sel.target.index:', sel.target.index)
                ind = sel.target.index
                annotationDict = self.getAnnotation(ind)
                myText = ''
                for k,v in annotationDict.items():
                    myText += f'{k}: {v}\n'
                sel.annotation.set_text(myText)

        #
        #self.mySetStatusBar(warningStr)

        self.canvas.axes.spines['right'].set_visible(False)
        self.canvas.axes.spines['top'].set_visible(False)

        if not stateDict['showLegend']:
            self.canvas.axes.legend().remove()

        #print('myUpdate() self.plotSize:', self.plotSize)
        self.canvas.axes.set_xlabel(xStatHuman)
        self.canvas.axes.set_ylabel(yStatHuman)
        '''
        if self.plotSize == 'paper':
            fontsize = 10
            self.canvas.axes[0].set_xlabel(xStatHuman, fontsize=fontsize)
            self.canvas.axes[0].set_ylabel(yStatHuman, fontsize=fontsize)
        else:
            self.canvas.axes[0].set_xlabel(xStatHuman)
            self.canvas.axes[0].set_ylabel(yStatHuman)
        '''

        # subplots_adjust
        #self.fig.canvas.draw_idle()
        self.fig.canvas.draw()


class bScatterPlotMainWindow(QtWidgets.QMainWindow):
    #send_fig = QtCore.pyqtSignal(str)
    signalStateChange = QtCore.Signal(object)
    signalSelectFromPlot = QtCore.Signal(object)
    signalCancelSelection = QtCore.Signal()
    signalMeanModelChange = QtCore.Signal(object)
    signal_xModelChange = QtCore.Signal(object)
    signal_yModelChange = QtCore.Signal(object)

    def __init__(self, path,
                    categoricalList, hueTypes, analysisName, sortOrder=None,
                    statListDict=None,
                    interfaceDefaults=None,
                    masterDf=None,
                    parent=None):
        """
        path: full path to .csv file generated with reanalyze
        categoricalList: specify columns that are categorical
            would just like to use 'if column is string' but sometimes number like 1/2/3 need to be categorical
        hueTypes:
        analysisName: column used for group by
        sortOrder:
        statListDict: dict where keys are human readable stat names
                    that map onto 'yStat' to specify column in csv
        interfaceDefaults: specify key/value in dict to set state of interface popups/etc
        masterDf: if not none then use it (rather than loading csv path)
                    used by main sanpy interface
        parent: not used, parent sanpy app

        todo: make pure text columns categorical
        todo: remove analysisName and add as 'groupby' popup from categorical columns
            depending on 'analysisName' popup, group differently in getMeanDf
        """
        super(bScatterPlotMainWindow, self).__init__(parent)

        #self.keepCheckBoxDelegate = myCheckBoxDelegate(None)

        if interfaceDefaults is None:
            interfaceDefaults = myInterfaceDefaults
        self.shortcut = QtWidgets.QShortcut(QtGui.QKeySequence("Ctrl+w"), self)
        self.shortcut.activated.connect(self.myCloseAction)

        self.statusBar = QtWidgets.QStatusBar()
        self.setStatusBar(self.statusBar)

        self.mplCursorHover = None

        self.buildMenus()

        # assigns self.masterDf
        # if masterDf is not None will use masterDf rather than loading path
        self.loadPath(path, masterDf=masterDf, categoricalList=categoricalList)

        # statListDict is a dict with key=humanstat name and yStat=column name in csv
        #self.statListDict = sanpy.bAnalysisUtil.getStatList()
        # 20210305 done in loadPath()
        if statListDict is not None:
            self.statListDict = statListDict
            # append all categorical
            if categoricalList is not None:
                for categorical in categoricalList:
                    self.statListDict[categorical] = {'yStat':categorical}
        '''
        if statListDict is None:
            # build statListDict from columns of csv path
            self.statListDict = {}
            for colStr in self.masterDfColumns:
                self.statListDict[colStr] = {'yStat': colStr}
        else:
            self.statListDict = statListDict
        '''

        #categoricalList = ['Condition', 'Sex', 'Region', 'File Number', 'File Name']
        # 20210305 done in load path
        '''
        for categorical in categoricalList:
            self.statListDict[categorical] = {'yStat': categorical}
        # this was originally in self.load() ???
        self.masterCatColumns = categoricalList
        '''

        '''
        self.stateDict = {}
        self.stateDict['groupByColumnName'] = analysisName
        self.stateDict['sortOrder'] = sortOrder
        self.stateDict['hue'] = hue
        self.stateDict['darkTheme'] = True
        self.stateDict['doKDE'] = False
        self.stateDict['doHover'] = False
        self.stateDict['plotSize'] = 'paper'
        self.stateDict['showLegend'] = True
        self.stateDict['plotType'] = 'Scatter Plot'
        self.stateDict['dataType'] = 'File Mean'
        '''

        # unique identifyer to group by
        # for sanpy this is 'analysisName', for bImPy this is xxx
        self.groupByColumnName = analysisName
        self.sortOrder = sortOrder

        # 20210112 moved up
        #self.loadPath(path)

        self.whatWeArePlotting = None # return from sns scatter plot (all plots)
        self.scatterPlotSelection = None
        self.xDf = None #
        self.yDf = None #
        self.plotDF = None # df we are plotting (can be same as mean yDf)
                            # use this to get row on self.onPick
        self.plotStatx = None
        self.plotStaty = None

        #self.main_widget = QtWidgets.QWidget(self)

        # this is causing tons of problems/crashes
        if interfaceDefaults['Hue'] is not None:
            self.hue = interfaceDefaults['Hue']
        else:
            self.hue = 'None' #self.hueTypes[0]

        self.darkTheme = False
        self.doKDE = False # fits on histogram plots
        self.showMplToolbar = False # fits on histogram plots

        self.doHover = False

        self.plotSizeList = ['paper', 'talk', 'poster']
        self.plotSize = 'paper'
        self.plotLayoutList = ['1x', '1x2', '2x1', '2x2']
        self.plotLayoutType = '2x2'
        self.updatePlot = None # which plot to update 1,2,3,4

        self.showLegend = False
        self.plotType = 'Scatter Plot'

        self.dataTypes = ['All Spikes', 'File Mean']
        self.dataType = 'File Mean' # (Raw, File Mean)

        self.buildUI(interfaceDefaults=interfaceDefaults)

        #bar = self.menuBar()
        #file = bar.addMenu("Load")

        self.show()

        # self.updatePlotSize() # calls update2()
        self.update2()

    def buildUI(self, interfaceDefaults):
        # if self.darkTheme:
        #     plt.style.use('dark_background')
        if self.darkTheme:
            plt.style.use('dark_background')
        else:
            plt.rcParams.update(plt.rcParamsDefault)

        # HBox for control and plot grid
        self.hBoxLayout = QtWidgets.QHBoxLayout(self)

        # this is confusing, beacaue we are a QMainWindow
        # we need to create a central widget, set its layout
        # and then set the central widget of self (QMainWindow)
        centralWidget = QtWidgets.QWidget()
        centralWidget.setLayout(self.hBoxLayout)
        self.setCentralWidget(centralWidget)

        # switch to 2x2
        '''
        # allow 1-4 plots
        tmpNumRow = 1
        tmpNumCol = 1
        numPlot = tmpNumRow * tmpNumCol

        self.fig = Figure()
        self.fig.tight_layout()
        # don't use subplot, each plot will be its own widget using myMplCanvas
        self.axes = [self.fig.add_subplot(tmpNumRow, tmpNumCol, i+1) for i in range(numPlot)]
        self.canvas = FigureCanvas(self.fig) # what is added to pyqt interface
        self.cid = self.canvas.mpl_connect('pick_event', self.on_pick_event)
        # maybe use
        #self.toolbar = NavigationToolbar(self.canvas, self)
        self.mplToolbar = NavigationToolbar2QT(self.canvas, self.canvas) # params are (canvas, parent)

        self.canvas.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                  QtWidgets.QSizePolicy.Expanding)
        self.canvas.updateGeometry()
        '''

        # to hold popups
        self.layout = QtWidgets.QGridLayout()
        self.hBoxLayout.addLayout(self.layout)

        def _defaultDropdownIdx(keyList, name):
            # if name not in keys, return 0
            theRet = 0
            try:
                theRet = keyList.index(name)
            except (ValueError) as e:
                print(f'WARNING: _defaultDropdownIdx() did not find "{name}"" in keyList: {keyList}')
                theRet = 0
            return theRet

        keys = list(self.statListDict.keys())
        #keys += self.hueTypes
        self.xDropdown = QtWidgets.QComboBox()
        self.xDropdown.addItems(keys)
        if interfaceDefaults['X Statistic'] is not None:
            defaultIdx = _defaultDropdownIdx(keys, interfaceDefaults['X Statistic'])
        else:
            defaultIdx = 0
        self.xDropdown.setCurrentIndex(defaultIdx)

        self.yDropdown = QtWidgets.QComboBox()
        self.yDropdown.addItems(keys)
        if interfaceDefaults['Y Statistic'] is not None:
            defaultIdx = _defaultDropdownIdx(keys, interfaceDefaults['Y Statistic'])
        else:
            defaultIdx = 0
        self.yDropdown.setCurrentIndex(defaultIdx)

        #
        # hue, to control colors in plot
        hueList = ['None'] + self.hueTypes # prepend 'None'
        self.hueDropdown = QtWidgets.QComboBox()
        self.hueDropdown.addItems(hueList)
        if interfaceDefaults['Hue'] is not None:
            defaultIdx = _defaultDropdownIdx(hueList, interfaceDefaults['Hue'])
        else:
            defaultIdx = 0
        self.hueDropdown.setCurrentIndex(defaultIdx) # 1 because we pre-pended 'None'
        self.hueDropdown.currentIndexChanged.connect(self.updateHue)

        #
        # group by, to control grouping in table
        # todo: get 'None' implemented
        #groupByList = ['None'] + self.hueTypes # prepend 'None'
        groupByList = self.hueTypes
        self.groupByDropdown = QtWidgets.QComboBox()
        self.groupByDropdown.addItems(groupByList)
        if interfaceDefaults['Group By'] is not None:
            defaultIdx = _defaultDropdownIdx(groupByList, interfaceDefaults['Group By'])
        else:
            defaultIdx = 0
        self.groupByDropdown.setCurrentIndex(defaultIdx) # 1 because we pre-pended 'None'
        self.groupByDropdown.currentIndexChanged.connect(self.updateGroupBy)

        # color
        #colorTypes = ['Region', 'Sex', 'Condition', 'File Number'] #, 'File Name'] #, 'None']
        #self.color = 'Region'
        '''
        self.color = self.colorTypes[0]
        self.colorDropdown = QtWidgets.QComboBox()
        self.colorDropdown.addItems(self.colorTypes)
        self.colorDropdown.setCurrentIndex(0)
        self.colorDropdown.currentIndexChanged.connect(self.updateColor)
        '''

        self.typeDropdown = QtWidgets.QComboBox()
        self.typeDropdown.addItems(['Scatter Plot', 'Scatter + Raw + Mean', 'Regression Plot', 'Violin Plot', 'Box Plot', 'Raw + Mean Plot', 'Histogram', 'Cumulative Histogram'])
        self.typeDropdown.setCurrentIndex(0)

        self.dataTypeDropdown = QtWidgets.QComboBox()
        self.dataTypeDropdown.setToolTip('All Spikes is all spikes \n File Mean is the mean within each analysis file')
        self.dataTypeDropdown.addItems(self.dataTypes)
        self.dataTypeDropdown.setCurrentIndex(1)

        self.xDropdown.currentIndexChanged.connect(self.update2)
        self.yDropdown.currentIndexChanged.connect(self.update2)
        self.typeDropdown.currentIndexChanged.connect(self.updatePlotType)
        self.dataTypeDropdown.currentIndexChanged.connect(self.updateDataType)

        showLegendCheckBox = QtWidgets.QCheckBox('Legend')
        showLegendCheckBox.setChecked(self.showLegend)
        showLegendCheckBox.stateChanged.connect(self.setShowLegend)

        # swap the sort order
        swapSortButton = QtWidgets.QPushButton('Swap Sort Order')
        swapSortButton.clicked.connect(self.setSwapSort)
        # works fine
        '''
        darkThemeCheckBox = QtWidgets.QCheckBox('Dark Theme')
        darkThemeCheckBox.setChecked(self.darkTheme)
        darkThemeCheckBox.stateChanged.connect(self.setTheme)
        darkThemeCheckBox.setDisabled(True)
        darkThemeCheckBox.setEnabled(False)
        '''

        '''
        kdeCheckBox = QtWidgets.QCheckBox('kde (hist)')
        kdeCheckBox.setChecked(self.doKDE)
        kdeCheckBox.stateChanged.connect(self.setKDE)
        '''
        mplToolbar = QtWidgets.QCheckBox('Toolbar')
        mplToolbar.setChecked(self.showMplToolbar)
        mplToolbar.stateChanged.connect(self.setMplToolbar)

        hoverCheckbox = QtWidgets.QCheckBox('Hover Info')
        hoverCheckbox.setChecked(self.doHover)
        hoverCheckbox.stateChanged.connect(self.setHover)

        # work fine
        '''
        self.plotSizeDropdown = QtWidgets.QComboBox()
        self.plotSizeDropdown.setToolTip('Set size of fonts for paper, talk, or poster')
        self.plotSizeDropdown.addItems(self.plotSizeList)
        self.plotSizeDropdown.setCurrentIndex(0) # paper
        self.plotSizeDropdown.currentIndexChanged.connect(self.updatePlotSize)
        self.plotSizeDropdown.setDisabled(True)
        '''

        # works fine
        '''
        self.plotLayoutDropdown = QtWidgets.QComboBox()
        self.plotLayoutDropdown.setToolTip('1, 2, or 4 plots')
        self.plotLayoutDropdown.addItems(self.plotLayoutList)
        self.plotLayoutDropdown.setCurrentIndex(3) #
        self.plotLayoutDropdown.currentIndexChanged.connect(self.updatePlotLayout)
        self.plotLayoutDropdown.setDisabled(True)
        '''

        # not needed beccause we now clik plot to show red square
        '''
        self.plotUpdateList = ['1', '2', '3', '4']
        self.plotUpdateDropdown = QtWidgets.QComboBox()
        self.plotUpdateDropdown.setToolTip('Which plot to update')
        self.plotUpdateDropdown.addItems(self.plotUpdateList)
        self.plotUpdateDropdown.setCurrentIndex(0) # paper
        self.plotUpdateDropdown.currentIndexChanged.connect(self.updatePlotUpdate)
        '''

        self.layout.addWidget(QtWidgets.QLabel("X Statistic"), 0, 0)
        self.layout.addWidget(self.xDropdown, 0, 1)
        self.layout.addWidget(QtWidgets.QLabel("Y Statistic"), 1, 0)
        self.layout.addWidget(self.yDropdown, 1, 1)
        self.layout.addWidget(QtWidgets.QLabel("Hue"), 2, 0)
        self.layout.addWidget(self.hueDropdown, 2, 1)
        self.layout.addWidget(QtWidgets.QLabel("Group By"), 3, 0)
        self.layout.addWidget(self.groupByDropdown, 3, 1)
        #self.layout.addWidget(QtWidgets.QLabel("Color"), 3, 0)
        #self.layout.addWidget(self.colorDropdown, 3, 1)
        #
        self.layout.addWidget(QtWidgets.QLabel("Plot Type"), 0, 2)
        self.layout.addWidget(self.typeDropdown, 0, 3)
        self.layout.addWidget(QtWidgets.QLabel("Data Type"), 1, 2)
        self.layout.addWidget(self.dataTypeDropdown, 1, 3)
        self.layout.addWidget(showLegendCheckBox, 2, 2)
        # works fine
        #self.layout.addWidget(darkThemeCheckBox, 2, 3)
        self.layout.addWidget(swapSortButton, 2, 3)
        self.layout.addWidget(mplToolbar, 3, 2)
        self.layout.addWidget(hoverCheckbox, 3, 3)
        # works fine
        #self.layout.addWidget(QtWidgets.QLabel("Plot Size"), 4, 2)
        #self.layout.addWidget(self.plotSizeDropdown, 4, 3)
        # works fine
        #self.layout.addWidget(QtWidgets.QLabel("Plot Layout"), 4, 0) # out of order
        #self.layout.addWidget(self.plotLayoutDropdown, 4, 1)

        #self.layout.addWidget(QtWidgets.QLabel("Update Plot"), 5, 0) # out of order
        #self.layout.addWidget(self.plotUpdateDropdown, 5, 1)

        nextRow = 5 # for text table
        rowSpan = 1
        colSpan = 4
        #self.layout.addWidget(self.canvas, 3, 0, rowSpan, colSpan)

        #
        # grid of plots (myMplCanvas)
        self.plotLayout = QtWidgets.QGridLayout()
        self.plotLayout.setContentsMargins(0,0,0,0)
        self.plotLayout.setHorizontalSpacing(0)
        self.plotLayout.setVerticalSpacing(0)

        self.myPlotCanvasList = [None] * 4

        self.updatePlotLayoutGrid()

        # append
        self.hBoxLayout.addLayout(self.plotLayout)

        '''
        self.myToolbar = QtWidgets.QToolBar()
        self.myToolbar.setFloatable(True)
        self.myToolbar.setMovable(True)
        self.tmpToolbarAction = self.myToolbar.addWidget(self.canvas)
        self.addToolBar(QtCore.Qt.RightToolBarArea, self.myToolbar)
        '''

        tabwidget = QtWidgets.QTabWidget()

        # hold main df database (imutable except for 'include'
        self.rawTableView = myTableView('All Spikes')
        self.rawTableView.clicked.connect(self.slotTableViewClicked)
        self.signalCancelSelection.connect(self.rawTableView.clearSelection)
        self.signalSelectFromPlot.connect(self.rawTableView.slotSelectRow)
        # never switch rawTableView model, dust update column 'include'
        #self.signalMeanModelChange.connect(self.rawTableView.slotSwitchTableModel)
        self.rawTableView.slotSwitchTableDf(self.masterDf)

        # self.xDf
        self.xTableView = myTableView('File Mean')
        self.xTableView.clicked.connect(self.slotTableViewClicked)
        self.signalCancelSelection.connect(self.xTableView.clearSelection)
        self.signalSelectFromPlot.connect(self.xTableView.slotSelectRow)
        # never switch rawTableView model, dust update column 'include'
        #self.signalMeanModelChange.connect(self.rawTableView.slotSwitchTableModel)
        self.xTableView.slotSwitchTableDf(self.xDf)

        # self.yDf
        self.yTableView = myTableView('File Mean')
        self.yTableView.clicked.connect(self.slotTableViewClicked)
        self.signalCancelSelection.connect(self.yTableView.clearSelection)
        self.signalSelectFromPlot.connect(self.yTableView.slotSelectRow)
        # never switch rawTableView model, dust update column 'include'
        #self.signalMeanModelChange.connect(self.rawTableView.slotSwitchTableModel)
        self.yTableView.slotSwitchTableDf(self.yDf)

        # todo: make this mean across yDf with stat test
        # mean table with pandas dataframe
        '''
        self.tableView = myTableView('Not Sure')
        self.tableView.clicked.connect(self.slotTableViewClicked)
        self.signalCancelSelection.connect(self.tableView.clearSelection)
        self.signalSelectFromPlot.connect(self.tableView.slotSelectRow)
        self.signalMeanModelChange.connect(self.tableView.slotSwitchTableModel)
        '''

        #
        #self.layout.addWidget(self.tableView, nextRow, 0, rowSpan, colSpan)
        tabwidget.addTab(self.rawTableView, "Raw")
        #tabwidget.addTab(self.tableView, "Mean")
        tabwidget.addTab(self.xTableView, "X-Stats")
        tabwidget.addTab(self.yTableView, "Y-Stats")
        self.layout.addWidget(tabwidget, nextRow, 0, rowSpan, colSpan)

    def updatePlotLayoutGrid(self):
        """
        use this to switch between (1x, 1x2, 2x1, 2x2)
        """

        print('updatePlotLayoutGrid()')
        plotLayoutType = self.plotLayoutType # 1x, 1x2, 2x1, 2x2
        if plotLayoutType == '1x':
            numPlots = 1
        elif plotLayoutType == '1x2':
            numPlots = 2
        elif plotLayoutType == '2x1':
            numPlots = 2
        elif plotLayoutType == '2x2':
            numPlots = 4

        # remove all widgets from self.plotLayout
        n = self.plotLayout.count()
        for i in range(n):
            item = self.plotLayout.itemAt(i)
            if item is None:
                print('  warning: updatePlotLayoutGrid() got None item at step', i)
                continue
            widget = item.widget()
            print('  updatePlotLayoutGrid() removing i:', i, 'item:', type(item))
            self.plotLayout.removeWidget(widget)
            #self.plotLayout.removeItem(item)

        state = self.getState()
        for i in range(numPlots):
            if i==0:
                row = 0
                col = 0
            elif i==1:
                if plotLayoutType == '1x2':
                    row = 0
                    col = 1
                elif plotLayoutType == '2x1':
                    row = 1
                    col = 0
                elif plotLayoutType == '2x2':
                    row = 0
                    col = 1
            elif i==2:
                row = 1
                col = 0
            elif i==3:
                row = 1
                col = 1
            #
            oneCanvas = myMplCanvas(plotNumber=i)
            oneCanvas.myUpdate(state) # initial plot
            self.signalCancelSelection.connect(oneCanvas.slotCancelSelection)
            self.myPlotCanvasList[i] = oneCanvas

            #
            self.plotLayout.addWidget(oneCanvas, row, col)

        # connect each canvas to all other canvas
        for i in range(numPlots):
            iCanvas = self.myPlotCanvasList[i]
            iCanvas.signalSelectSquare.connect(self.slotSelectSquare)
            iCanvas.signalSelectFromPlot.connect(self.slotSelectFromPlot)
            for j in range(numPlots):
                #if i==j:
                #    continue
                jCanvas = self.myPlotCanvasList[j]
                iCanvas.signalSelectFromPlot.connect(jCanvas.slotSelectInd)
                iCanvas.signalSelectSquare.connect(jCanvas.slotSelectSquare)

        #
        # select the firsr plot
        self.myPlotCanvasList[0].signalSelectSquare.emit(0, None) #slotSelectSquare(0)

    def keyPressEvent(self, event):
        print('keyPressEvent()')
        if event.key() == QtCore.Qt.Key_Escape:
            self.cancelSelection()
        elif (event.type() == QtCore.QEvent.KeyPress and
                        event.matches(QtGui.QKeySequence.Copy)):
            self.copySelection2()
        #
        event.accept()

    def copySelection2(self):
        dfCopy = None
        if self.xDf is not None:
            dfCopy = self.xDf.copy()
            #self.xDf.to_clipboard(sep='\t', index=False)
            #print('Copied to clipboard')
            #print(self.xDf)
        if self.yDf is not None:
            if dfCopy is None:
                dfCopy = self.yDf.copy()
            else:
                # append row then yDf
                dfCopy.append(pd.Series(), ignore_index=True)
                dfCopy.append(self.yDf, ignore_index=True)
            #self.yDf.to_clipboard(sep='\t', index=False)
            #print('Copied to clipboard')
            #print(self.yDf)
        #
        if dfCopy is not None:
            dfCopy.to_clipboard(sep='\t', index=False)
            print(dfCopy)

    """
    def eventFilter(self, source, event):
        if (event.type() == QtCore.QEvent.KeyPress and
                        event.matches(QtGui.QKeySequence.Copy)):
            self.copySelection2()
            return True
        return super(bScatterPlotMainWindow, self).eventFilter(source, event)
    """

    def cancelSelection(self):
        self.signalCancelSelection.emit()

    def _switchTableModel(self, newModel):
        """
        switch model for mean table (Self.xxx)
        """
        #print('bScatterPlotMainWindow._switchTableModel()')
        self.signalMeanModelChange.emit(newModel)

    def getState(self):
        """
        query all controls and create dict with state

        used by myMplCanvas.update()
        """
        stateDict = {}

        stateDict['dataType'] = self.dataType # ['All spikes', 'File Mean']
        stateDict['groupByColumnName'] = self.groupByColumnName

        plotType = self.plotType
        stateDict['plotType'] = plotType

        xStatHuman = self.xDropdown.currentText()
        yStatHuman = self.yDropdown.currentText()
        stateDict['xStatHuman'] = xStatHuman
        stateDict['yStatHuman'] = yStatHuman
        xStat = self.statListDict[xStatHuman]['yStat'] # statListDict always used yStat
        yStat = self.statListDict[yStatHuman]['yStat']
        stateDict['xStat'] = xStat
        stateDict['yStat'] = yStat
        stateDict['xIsCategorical'] = pd.api.types.is_string_dtype(self.masterDf[xStat].dtype)
        stateDict['yIsCategorical'] = pd.api.types.is_string_dtype(self.masterDf[yStat].dtype)

        if self.hue == 'None':
            # special case, we do not want None in self.statNameMap
            hue = None
        else:
            hue = self.hue
        stateDict['hue'] = hue

        stateDict['darkTheme'] = self.darkTheme
        stateDict['showLegend'] = self.showLegend
        stateDict['doHover'] = self.doHover
        stateDict['showMplToolbar'] = self.showMplToolbar

        if self.dataType == 'All Spikes':
            meanDf = self.masterDf
            if self.sortOrder is not None:
                # need to sort so onPick works
                print('  ', '(1) sorting by self.sortOrder:', self.sortOrder)
                meanDf = meanDf.sort_values(self.sortOrder)

            # remove rows that have nan in our x or y stat
            meanDf = meanDf[~meanDf[xStat].isnull()]
            meanDf = meanDf[~meanDf[yStat].isnull()]
            meanDf = meanDf.reset_index()

        elif self.dataType == 'File Mean':
            meanDf = self.getMeanDf(xStat, yStat)
        else:
            print('error in self.dataType:', self.dataType)

        stateDict['masterDf'] = self.masterDf
        stateDict['meanDf'] = meanDf
        return stateDict

    def myCloseAction(self):
        print('myCloseAction()')

    def mySetStatusBar(self, text):
        self.statusBar.showMessage(text) #,2000)

    '''
    def copyTable(self):
        headerList = []
        for i in self.tableView.model().columnCount():
            headers.append(self.tableView.model().headerData(i, QtCore.Qt.Horizontal).toString()
        print('copyTable()')
        print('  headers:', headers)
        m = self.tableView.rowCount()
        n = self.tableView.columnCount()
        table = [[''] * n for x in range(m+1)]
        #for i in m:
    '''

    '''
    def copySelection(self):
        #self.copyTable()

        selection = self.tableView.selectedIndexes()
        if selection:
            rows = sorted(index.row() for index in selection)
            columns = sorted(index.column() for index in selection)
            rowcount = rows[-1] - rows[0] + 1
            colcount = columns[-1] - columns[0] + 1
            table = [[''] * colcount for _ in range(rowcount)]
            for index in selection:
                row = index.row() - rows[0]
                column = index.column() - columns[0]
                table[row][column] = index.data()
            stream = io.StringIO()
            csv.writer(stream).writerows(table)
            QtWidgets.QApplication.clipboard().setText(stream.getvalue())
    '''

    def buildMenus(self):
        loadAction = QtWidgets.QAction('Load database.xlsx', self)
        loadAction.triggered.connect(self.loadPathMenuAction)

        menubar = self.menuBar()
        fileMenu = menubar.addMenu('&File')
        fileMenu.addAction(loadAction)

    def loadPathMenuAction(self):
        """
        prompt user for database.xlsx
        run reanalyze.py on that xlsx database
        load resultant _master.csv
        """
        print('loadPathMenuAction')

    def loadPath(self, path, masterDf=None, categoricalList=None):
        """
        path: full path to .csv file generated with reanalyze.py
        """
        #path = '/Users/cudmore/data/laura-ephys/Superior vs Inferior database_master.csv'
        if masterDf is not None:
            self.masterDf = masterDf
        else:
            if path.endswith('.csv'):
                self.masterDf = pd.read_csv(path, header=0) #, dtype={'ABF File': str})
            elif path.endswith('.xls'):
                self.masterDf = pd.read_excel(path, header=0) #, dtype={'ABF File': str})
            elif path.endswith('.xlsx'):
                self.masterDf = pd.read_excel(path, header=0, engine='openpyxl') #, dtype={'ABF File': str})
            else:
                print('error: file type not supported. Expecting csv/xls/xlsx. Path:', path)

        # 20210426
        # combine some columns into new encoding
        # like, 'male+superior' as 'ms'
        # moved to reanalyze
        # want to move to reanalyze
        # but my base analysis does not have region/sex/condition etc
        tmpNewCol = 'RegSex'
        self.masterDf[tmpNewCol] = ''
        for tmpRegion in ['Superior', 'Inferior']:
            for tmpSex in ['Male', 'Female']:
                newEncoding = tmpRegion[0] + tmpSex[0]
                regSex = self.masterDf[ (self.masterDf['Region']==tmpRegion) & (self.masterDf['Sex']==tmpSex)]
                regSex = (self.masterDf['Region']==tmpRegion) & (self.masterDf['Sex']==tmpSex)
                print('newEncoding:', newEncoding, 'regSex:', regSex.shape)
                self.masterDf.loc[regSex, tmpNewCol] = newEncoding

        #sys.exit()

        self.masterDfColumns = self.masterDf.columns.to_list()

        #
        # try and guess categorical columns
        # if int64 and unique()<20 then assume it is category (works for date yyyymmdd)
        #print('loadPath() dtypes:')
        #print(self.masterDf.dtypes)
        if categoricalList is None:
            categoricalList = []
            for colStr, dtype in self.masterDf.dtypes.items():
                if dtype==object:
                    print('colStr:', colStr, 'is', dtype)
                    categoricalList.append(colStr)
                elif dtype==np.int64:
                    unique = self.masterDf[colStr].unique()
                    numUnique = len(unique)
                    print('colStr:', colStr, 'is', dtype, 'numUnique:', numUnique)
                    categoricalList.append(colStr)

        self.masterCatColumns = categoricalList
        self.hueTypes = categoricalList

        self.statListDict = {}
        for colStr in self.masterDfColumns:
            self.statListDict[colStr] = {'yStat': colStr}

        # not sure what this was for ???
        # 20210112, put back in if necc
        #self.masterCatColumns = ['Condition', 'File Number', 'Sex', 'Region', 'filename', 'analysisname']
        #self.masterCatColumns = self.categoricalList

        # todo: put this somewhere better
        self.setWindowTitle(path)

    def setSwapSort(self):
        """
        rotate the sort key to be able to sort by different keys, like 'sex' versus 'region'
        """
        #print('=== setSwapSort()')
        #print('  sortOrder was:', self.sortOrder)
        self.sortOrder = self.sortOrder[1:] + self.sortOrder[:1]
        #print('  sortOrder now:', self.sortOrder)
        self.mySetStatusBar(f'Sort order is now: {self.sortOrder}')

    def setShowLegend(self, state):
        print('setShowLegend() state:', state)
        self.showLegend = state
        self.updateGlobal()

    '''
    def setKDE(self, state):
        # only used in histograms
        self.doKDE = state
        self.updateGlobal()
    '''

    def setMplToolbar(self, state):
        # only used in histograms
        self.showMplToolbar = state
        self.updateGlobal()

    def setHover(self, state):
        # used in scatterplots and point plots
        self.doHover = state
        #self.signalStateChange.emit(self.getState())
        self.updateGlobal()

    def setTheme(self, state):
        print('setTheme() state:', state)

        self.darkTheme = state
        if self.darkTheme:
            plt.style.use('dark_background')
            #sns.set_context('talk')

        else:
            #print(plt.style.available)
            plt.rcParams.update(plt.rcParamsDefault)
            #sns.set_context('paper')

        #self.updateGlobal()
        n = len(self.myPlotCanvasList)
        for i in range(n):
            if self.myPlotCanvasList[i] is not None:
                self.myPlotCanvasList[i].updateTheme()

        ###
        return
        ###

        ###
        ###
        # remove
        #self.plotVBoxLayout.removeWidget(self.toolbar)
        # need this !!! removeWidget does not work
        #self.myToolbar.removeAction(self.tmpToolbarAction)
        self.hBoxLayout.removeWidget(self.canvas)
        self.canvas.setParent(None)

        self.fig = None # ???
        self.mplToolbar = None
        self.canvas = None

        self.fig = Figure()
        self.canvas = FigureCanvas(self.fig)
        tmpAx = self.fig.add_subplot(111) # self.ax1 not used
        self.axes = [tmpAx]
        self.cid = self.canvas.mpl_connect('pick_event', self.on_pick_event)
        # matplotlib navigation toolbar
        self.mplToolbar = NavigationToolbar2QT(self.canvas, self.canvas) # params are (canvas, parent)
        self.hBoxLayout.addWidget(self.canvas)

        #
        # add a second plot
        if self.canvas2 is not None:
            self.hBoxLayout.removeWidget(self.canvas2)
            self.canvas2.setParent(None)
        self.canvas2 = myMplCanvas()
        self.signalStateChange.connect(self.canvas2.myUpdate)
        tmpState = self.getState()
        self.canvas2.myUpdate(tmpState)
        self.hBoxLayout.addWidget(self.canvas2)

        # original plot
        self.update2()

    def updatePlotLayout(self):
        """
        set number of plots [1x, 1x2, 2x1, 2x2]
        """
        self.plotLayoutType = self.plotLayoutDropdown.currentText() # ['paper', 'poster', 'talk']

        self.updatePlotLayoutGrid()

        self.update2()

    def updatePlotSize(self):
        self.plotSize = self.plotSizeDropdown.currentText() # ['paper', 'poster', 'talk']
        sns.set_context(self.plotSize) #, font_scale=1.4)
        self.updateGlobal()

    def updateHue(self):
        hue = self.hueDropdown.currentText()
        self.hue = hue
        #self.signalStateChange.emit(self.getState())
        self.update2()

    def updateGroupBy(self):
        self.groupByColumnName = self.groupByDropdown.currentText()
        self.update2() # todo: don't update plot, update table

    '''
    def updateColor(self):
        color = self.colorDropdown.currentText()
        self.color = color
        self.update2()
    '''

    def updatePlotType(self):
        plotType = self.typeDropdown.currentText()
        self.plotType = plotType
        self.update2()

    def updateDataType(self):
        dataType = self.dataTypeDropdown.currentText()
        self.dataType = dataType
        self.update2()

    '''
    def old_on_pick_event(self, event):
        try:
            print('on_pick_event() event:', event)
            print('event.ind:', event.ind)

            if len(event.ind) < 1:
                return
            spikeNumber = event.ind[0]
            print('  selected:', spikeNumber)

            # propagate a signal to parent
            #self.myMainWindow.mySignal('select spike', data=spikeNumber)
            #self.selectSpike(spikeNumber)
        except (AttributeError) as e:
            pass
    '''

    def getMeanDf(self, xStat, yStat, verbose=False):
        # need to get all categorical columns from orig df
        # these do not change per file (sex, condition, region)
        print('=== getMeanDf() xStat:', xStat, 'yStat:', yStat)
        print('  self.groupByColumnName:', self.groupByColumnName)

        xIsCategorical = pd.api.types.is_string_dtype(self.masterDf[xStat].dtype)
        yIsCategorical = pd.api.types.is_string_dtype(self.masterDf[yStat].dtype)

        groupByNone = self.groupByColumnName == 'None'

        aggList = ['count', 'mean', 'std', 'sem', 'median']
        if xIsCategorical or groupByNone:
            self.xDf = None
        else:
            self.xDf = self.masterDf.groupby(self.groupByColumnName, as_index=False)[xStat].agg(aggList)
            self.xDf = self.xDf.reset_index()
            self.xDf.insert(1, 'stat', xStat) # column 1, in place

        if yIsCategorical or groupByNone:
            self.yDf = None
        else:
            self.yDf = self.masterDf.groupby(self.groupByColumnName, as_index=False)[yStat].agg(aggList)
            self.yDf = self.yDf.reset_index()
            self.yDf.insert(1, 'stat', yStat) # column 1, in place

        if xStat == yStat:
            groupList = [xStat]
        else:
            groupList = [xStat, yStat]
        if groupByNone:
            meanDf = None
        else:
            meanDf = self.masterDf.groupby(self.groupByColumnName, as_index=False)[groupList].mean()
            meanDf = meanDf.reset_index()

        #
        # 20210211 get median/std/sem/n
        # try and add median/std/sem/sem/n
        if 0:
            tmpDf = self.masterDf.groupby(self.groupByColumnName, as_index=False)[groupList].median()
            #print('tmpDf:', tmpDf)
            meanDf['median'] = tmpDf[self.groupByColumnName]

        '''
        for catName in self.masterCatColumns:
            #if catName == 'analysisname':
            if catName == self.groupByColumnName:
                # this is column we grouped by, already in meanDf
                continue
            meanDf[catName] = ''
        '''

        # for each row, update all categorical columns using self.masterCatColumns
        fileNameList = meanDf[self.groupByColumnName].unique()
        #print('  getMeanDf() updating all categorical columns for rows in fileNameList:')#, fileNameList)
        #print('    categorical columns are self.masterCatColumns:', self.masterCatColumns)
        for i, analysisname in enumerate(fileNameList):
            tmpDf = self.masterDf[ self.masterDf[self.groupByColumnName]==analysisname ]
            if len(tmpDf) == 0:
                print('  ERROR: got 0 length for analysisname:', analysisname)
                continue
            # need to limit this to pre-defined catecorical columns
            for catName in self.masterCatColumns:
                #if i==0:
                #    print('analysisname:', analysisname, 'catName:', catName)

                #if catName == self.groupByColumnName:
                #    # this is column we grouped by, already in meanDf
                #    continue

                # find value of catName column from 1st instance in masterDf
                # if more than one unique value, don't put into table
                numUnique = len(tmpDf[catName].unique())
                if numUnique == 1:
                    #print('    updating categorical colum with catName:', catName)
                    catValue = tmpDf[catName].iloc[0]

                    theseRows = (meanDf[self.groupByColumnName]==analysisname).tolist()
                    #if catName == self.groupByColumnName:
                    meanDf.loc[theseRows, catName] = catValue
                    if self.xDf is not None:
                        self.xDf.loc[theseRows, catName] = catValue
                    if self.yDf is not None:
                        self.yDf.loc[theseRows, catName] = catValue
                    #print('catName:', catName, 'catValue:', type(catValue), catValue)
                else:
                    print('warning catName:', catName, 'has', numUnique, 'unique values')
                    pass
                    #print(f'not adding {catName} to table, numUnique: {numUnique}')
        #
        # sort
        #meanDf = meanDf.sort_values(['Region', 'Sex', 'Condition'])
        if self.sortOrder is not None:
            print('  ', 'sorting by self.sortOrder:', self.sortOrder)
            try:
                meanDf = meanDf.sort_values(self.sortOrder)
                if self.xDf is not None:
                    self.xDf = self.xDf.sort_values(self.sortOrder)
                if self.yDf is not None:
                    self.yDf = self.yDf.sort_values(self.sortOrder)
            except (KeyError) as e:
                print('    ', 'sorting (2) failed with:', e)
        #
        meanDf['index'] = [x+1 for x in range(len(meanDf))]
        meanDf = meanDf.reset_index()
        if self.xDf is not None:
            self.xDf['index'] = [x+1 for x in range(len(self.xDf))]
            self.xDf = self.xDf.reset_index()
        if self.yDf is not None:
            self.yDf['index'] = [x+1 for x in range(len(self.yDf))]
            self.yDf = self.yDf.reset_index()

        # drop extra columns
        #modelMeanDf = meanDf.drop(['level_0'], axis=1)
        if self.xDf is not None and 'level_0' in self.xDf.columns:
            self.xDf = self.xDf.drop(['level_0'], axis=1)
        if self.yDf is not None and 'level_0' in self.yDf.columns:
            self.yDf = self.yDf.drop(['level_0'], axis=1)
        # we need to round for the table, but NOT the plot !!!!!
        #meanDf = meanDf.round(3)

        #
        if verbose:
            print('getMeanDf():')
            print(meanDf)
        #
        return meanDf

    def updateGlobal(self):
        """
        update all plots
        do not replot
        """
        state = self.getState()
        n = len(self.myPlotCanvasList)
        for i in range(n):
            if self.myPlotCanvasList[i] is not None:
                self.myPlotCanvasList[i].myUpdateGlobal(state)

    def update2(self):
        #self.updatePlot is the plot to update 1,2,3,4
        updateIndex = self.updatePlot
        if updateIndex is None:
            print('update2() found no plots to update with self.updatePlot:', self.updatePlot)
            self.mySetStatusBar('Please select a plot to update')
            return

        state = self.getState()

        # update table model
        meanDf = state['meanDf']
        #modelMeanDf = meanDf.drop(['level_0'], axis=1)
        #self.myModel = myPandasModel(meanDf) # todo: don't need self.myModel
        myModel = myPandasModel(meanDf) # todo: don't need self.myModel
        #print('calling _switchTableModel from update2() updateIndex:', updateIndex)
        self._switchTableModel(myModel)

        self.xTableView.slotSwitchTableDf(self.xDf)
        self.yTableView.slotSwitchTableDf(self.yDf)

        # update plots
        n = len(self.myPlotCanvasList)
        for i in range(n):
            if i==updateIndex:
                if self.myPlotCanvasList[i] is not None:
                    self.myPlotCanvasList[i].myUpdate(state)

    def old_getAnnotation(self, ind):
        # todo: replace with _getStatFromPlot

        if not np.issubdtype(ind, np.integer):
            print('getAnnotation() got bad ind:', ind, type(ind))
            return

        analysisName = self.plotDf.at[ind, self.groupByColumnName]
        index = self.plotDf.at[ind, 'index']
        try:
            region = self.plotDf.at[ind, 'Region'] # not all will have this
        except (KeyError) as e:
            region = 'n/a'
        xVal = self.plotDf.at[ind, self.plotStatx]
        yVal = self.plotDf.at[ind, self.plotStaty]

        returnDict = {
                    'index': index,
                    'analysisName': analysisName,
                    'region': region,
                    'xVal': xVal,
                    'yVal': yVal,
                    #'plotDf': self.plotDf, # potentially very big
            }
        return returnDict

    # see: https://stackoverflow.com/questions/7908636/possible-to-make-labels-appear-when-hovering-over-a-point-in-matplotlib
    """
    def old_onHover(self, event):
        print('onHover:', type(event), 'inaxes:', event.inaxes)
        if event.inaxes == self.axes[0]:
            print('  in plotted axes')
        else:
            print('  not in plotted axes:', self.axes[0])

        print('  whatWeArePlotting:', type(self.whatWeArePlotting))
        cont, ind = self.whatWeArePlotting.contains(event)
        print('  cont:', cont)
        print('  ind:', ind)
        '''
        ind = event.ind # ind is a list []
        ind = ind[0]
        self._getStatFromPlot(ind)
        '''
    """

    '''
    def old_getStatFromPlot(self, ind):
        """
        get stat from self.plotDf from connected click/hover
        """
        analysisName = self.plotDf.at[ind, self.groupByColumnName]
        index = self.plotDf.at[ind, 'index']
        region = self.plotDf.at[ind, 'region']
        xVal = self.plotDf.at[ind, self.plotStatx]
        yVal = self.plotDf.at[ind, self.plotStaty]
        print(f'index:{index}, analysisName:{analysisName}, region:{region}, {self.plotStatx}:{xVal}, {self.plotStaty}:{yVal}')
    '''

    def slotSelectFromPlot(self, selectDict):
        """
        A point in one of plots was selected
        pass this to parent app
        """
        print('bScatterPlotMainWindow.slotSelectFromPlot() ', selectDict)

    def slotSelectSquare(self, plotNumber, stateDict):
        print('bScatterPlotMainWindow.slotSelectSquare()', plotNumber)
        #print('  stateDict:', stateDict)
        print('  todo: fill in all interface using stateDict')
        self.updatePlot = plotNumber

    def slotTableViewClicked(self, clickedIndex):
        """
        Respond to signal self.tableView.clicked
        clickedIndex: PyQt5.QtCore.QModelIndex
        """
        row=clickedIndex.row()
        model=clickedIndex.model()
        print('bScatterPlotMainWindow.slotTableViewClicked() row:', row, 'clickedIndex:', clickedIndex)

        # select in plot
        #self._selectInd(row) # !!!! visually, index start at 1

if __name__ == '__main__':
    """
    20210112, extending this to work with any csv. Starting with nodes/edges from bimpy
    """

    # todo: using 'analysisname' for group by, I think I can also use 'File Number'
    statListDict = None # list of dict mapping human readbale to column names
    masterDf = None
    interfaceDefaults = None

    # machine learning db
    if 0:
        # this is from mac laptop
        #path = '/Users/cudmore/data/laura-ephys/Superior vs Inferior database_master.csv'
        path = '/Users/cudmore/data/laura-ephys/SANdatabaseForMachineLearning.xlsx'
        analysisName = 'File Number'
        #statListDict = None #sanpy.bAnalysisUtil.getStatList()
        categoricalList = ['LOCATION', 'SEX', 'File Number']#, 'File Name']
        hueTypes = ['LOCATION', 'SEX', 'File Number'] #, 'File Name'] #, 'None']
        sortOrder = ['LOCATION', 'SEX', 'File Number']

    # sanpy database
    if 1:
        #import sanpy
        #sys.path.append(os.path.join(os.path.dirname(sys.path[0]),'sanpy'))
        #import bAnalysisUtil
        #statListDict = bAnalysisUtil.statList
        import statlist
        statListDict = statlist.statList

        # this is from mac laptop
        #path = '/Users/cudmore/data/laura-ephys/Superior vs Inferior database_master.csv'
        path = '../examples/Superior vs Inferior database_master.csv'
        path = '/Users/cudmore/data/laura-ephys/Superior_Inferior_database_master_jan25.csv'
        path = '/Users/cudmore/data/laura-ephys/Superior vs Inferior database_master.csv'
        path = '/Users/cudmore/data/laura-ephys/Superior vs Inferior database_master_20210402.csv'

        #path = 'data/Superior vs Inferior database_master_20210402.csv'
        path = 'data/Superior vs Inferior database_master_20210402.csv'
        #path = '/Users/cudmore/data/laura-ephys/Superior_Inferior_database_master_jan25.csv'
        path = 'data/Superior vs Inferior database_13_Feb_master.csv'
        analysisName = 'analysisname'
        #statListDict = None #sanpy.bAnalysisUtil.getStatList()
        categoricalList = ['include', 'Condition', 'Region', 'Sex', 'RegSex', 'File Number', 'analysisname']#, 'File Name']
        hueTypes = ['Region', 'Sex', 'RegSex', 'Condition', 'File Number', 'analysisname'] #, 'File Name'] #, 'None']
        sortOrder = ['Region', 'Sex', 'Condition']

        interfaceDefaults = {'Y Statistic': 'Spike Frequency (Hz)',
                            'X Statistic': 'Region',
                            'Hue': 'Region',
                            'Group By': 'File Number'}
    # bimpy database
    if 0:
        path = '../examples/edges_db.csv'
        analysisName = 'fileNumber'
        categoricalList = ['san', 'region', 'path', 'file', 'fileNumber', 'nCon']
        hueTypes = categoricalList
        sortOrder = ['san', 'region']

    # dualAnalysis database
    if 0:
        # grab our list of dict mapping human readable to .csv column names
        sys.path.append(os.path.join(os.path.dirname(sys.path[0]),'sanpy'))
        import bAnalysisUtil
        statListDict = bAnalysisUtil.statList

        path = '/Users/cudmore/Sites/SanPy/examples/dual-analysis/dualAnalysis_final_db.csv'
        analysisName = 'fileNumber' # # rows in .xlsx database, one recording per row
        # trial is 1a/1b/1c... trial withing cellNumber
        categoricalList = ['include', 'region', 'fileNumber', 'cellNumber', 'trial', 'quality']
        hueTypes = categoricalList
        sortOrder = ['region']

    # sparkmaster lcr database
    if 0:
        path = '/Users/cudmore/Sites/SanPy/examples/dual-analysis/lcr-database.csv'
        analysisName = 'fileNumber' # # rows in .xlsx database, one recording per row
        # trial is 1a/1b/1c... trial withing cellNumber
        categoricalList = ['quality', 'region', 'fileNumber', 'dateFolder', 'tifFile']
        hueTypes = categoricalList
        sortOrder = ['region']

    # lcr/vm analysis using lcrPicker.py
    if 0:
        #basePath = '/Users/cudmore/Sites/SanPy/examples/dual-analysis/'
        #path = basePath + 'dual-data/20210115/20210115__0002_lcrPicker.csv'
        #path = basePath + 'dual-data/20210115/20210115__0001_lcrPicker.csv'

        # output of lcrPicker.py ... mergeDatabase()
        path = '/Users/cudmore/Sites/SanPy/examples/dual-analysis/lcrPicker-db.csv'
        categoricalList = None
        hueTypes = None
        analysisName= 'tifFile'
        sortOrder = None

    # merged sanpy+lcr pre spike slope
    # generated by dualAnalysis.py xxx()
    # usnig to compare lcr slope to edddr for fig 9
    if 0:
        path = '/Users/cudmore/Sites/SanPy/examples/dual-analysis/combined-sanpy-lcr-db.csv'
        statListDict = None
        categoricalList = None
        hueTypes = None
        analysisName= 'filename'
        sortOrder = None

    #
    app = QtWidgets.QApplication(sys.argv)

    ex = bScatterPlotMainWindow(path, categoricalList, hueTypes,
                    analysisName, sortOrder, statListDict=statListDict,
                    masterDf = masterDf,
                    interfaceDefaults = interfaceDefaults)

    sys.exit(app.exec_())



"""General purpose (numerical) utilities for spike detection and plotting.
"""

import numpy as np

from sanpy.sanpyLogger import get_logger
logger = get_logger(__name__)

def throwOutAboveBelow(vm, spikeTimes, spikeErrors, peakWindow_pnts, onlyPeaksAbove_mV=None, onlyPeaksBelow_mV=None):
    """
    Args:
        vm (np.ndarray):
        spikeTimes (list): list of spike times
        spikeErrors (list): list of error
    """
    newSpikeTimes = []
    newSpikeErrorList = []
    newSpikePeakPnt = []
    newSpikePeakVal = []
    for i, spikeTime in enumerate(spikeTimes):
        peakPnt = np.argmax(vm[spikeTime:spikeTime+peakWindow_pnts])
        peakPnt += spikeTime
        peakVal = np.max(vm[spikeTime:spikeTime+peakWindow_pnts])

        goodSpikeAbove = True
        if onlyPeaksAbove_mV is not None and (peakVal < onlyPeaksAbove_mV):
            goodSpikeAbove = False
        goodSpikeBelow = True
        if onlyPeaksBelow_mV is not None and (peakVal > onlyPeaksBelow_mV):
            goodSpikeBelow = False

        if goodSpikeAbove and goodSpikeBelow:
            newSpikeTimes.append(spikeTime)
            newSpikeErrorList.append(spikeErrors[i])
            newSpikePeakPnt.append(peakPnt)
            newSpikePeakVal.append(peakVal)
        else:
            #print('spikeDetect() peak height: rejecting spike', i, 'at pnt:', spikeTime, "dDict['onlyPeaksAbove_mV']:", dDict['onlyPeaksAbove_mV'])
            pass
    #
    return newSpikeTimes, newSpikeErrorList, newSpikePeakPnt, newSpikePeakVal

def getEddLines(ba):
    """Get lines representing linear fit of EDD rate.

    Args:
        ba (bAnalysis): bAnalysis object
    """
    logger.info(ba)

    x = []
    y = []
    if ba is None or ba.numSpikes == 0:
        return x, y

    #
    # these are getting for current sweep
    preLinearFitPnt0 = ba.getStat('preLinearFitPnt0')
    preLinearFitSec0 = [ba.fileLoader.pnt2Sec_(x) for x in preLinearFitPnt0]
    preLinearFitVal0 = ba.getStat('preLinearFitVal0')

    preLinearFitPnt1 = ba.getStat('preLinearFitPnt1')
    preLinearFitSec1 = [ba.fileLoader.pnt2Sec_(x) for x in preLinearFitPnt1]
    preLinearFitVal1 = ba.getStat('preLinearFitVal1')

    thisNumSpikes = len(preLinearFitPnt0)
    #for idx, spike in enumerate(range(ba.numSpikes)):
    for idx, spike in enumerate(range(thisNumSpikes)):
        #dx = preLinearFitSec1[idx] - preLinearFitSec0[idx]
        #dy = preLinearFitVal1[idx] - preLinearFitVal0[idx]

        # always plot edd rate line at constant length
        lineLengthSec = 0.1 #8  # TODO: make this a function of spike frequency?

        x.append(preLinearFitSec0[idx])
        x.append(preLinearFitSec1[idx] + lineLengthSec)
        x.append(np.nan)

        y.append(preLinearFitVal0[idx])
        y.append(preLinearFitVal1[idx] + lineLengthSec)
        y.append(np.nan)

    #logger.info(f'{len(x)} {len(y)}')

    return x, y

def getHalfWidthLines(t, v, spikeDictList):
    """Get x/y pair for plotting all half widths."""
    x = []
    y = []

    #if len(t.shape)>1 or len(v.shape)>1:
    #    logger.error('EXPECTING 1D, EXPAND THIS TO 2D SWEEPS')
    #    print(t.shape, v.shape)
    #    return x, y

    if t.shape != v.shape:
        logger.error(f't:{t.shape} and v:{v.shape} are not the same length')
        return x, y

    is2D = len(t.shape) > 1

    numPerSpike = 3  # rise/fall/nan
    #numSpikes = self.ba.numSpikes
    numSpikes = len(spikeDictList)
    logger.info(f'numSpikes:{numSpikes}')
    xyIdx = 0
    #for idx, spike in enumerate(self.ba.spikeDict):
    #spikeDictionaries = self.ba.getSpikeDictionaries(sweepNumber=self.sweepNumber)
    for idx, spike in enumerate(spikeDictList):
        sweep = spike['sweep']
        if idx ==0:
            # make x/y from first spike using halfHeights = [20,50,80,...]
            halfHeights = spike['halfHeights'] # will be same for all spike, like [20, 50, 80]
            numHalfHeights = len(halfHeights)
            # *numHalfHeights to account for rise/fall + padding nan
            x = [np.nan] * (numSpikes * numHalfHeights * numPerSpike)
            y = [np.nan] * (numSpikes * numHalfHeights * numPerSpike)
            #print('  len(x):', len(x), 'numHalfHeights:', numHalfHeights, 'numSpikes:', numSpikes, 'halfHeights:', halfHeights)

        if 'widths' not in spike.keys():
            logger.error(f'=== Did not find "widths" key in spike {idx}')

        for idx2, width in enumerate(spike['widths']):
            #halfHeight = width['halfHeight'] # [20,50,80]
            risingPnt = width['risingPnt']
            #risingVal = width['risingVal']
            fallingPnt = width['fallingPnt']
            #fallingVal = width['fallingVal']

            if risingPnt is None or fallingPnt is None:
                # half-height was not detected
                continue

            #risingSec = self.ba.pnt2Sec_(risingPnt)
            #fallingSec = self.ba.pnt2Sec_(fallingPnt)


            # if v or t are 2D, we have multiple sweeps
            # one sweep per column
            if is2D:
                risingVal = v[risingPnt,sweep]
                fallingVal = v[fallingPnt,sweep]
                risingSec = t[risingPnt,sweep]
                fallingSec = t[fallingPnt,sweep]
            else:
                risingVal = v[risingPnt]
                fallingVal = v[fallingPnt]
                risingSec = t[risingPnt]
                fallingSec = t[fallingPnt]

            #print(f'fallingVal:{fallingVal} {type(fallingVal)}')

            # x
            x[xyIdx] = risingSec
            x[xyIdx+1] = fallingSec
            x[xyIdx+2] = np.nan
            # y
            y[xyIdx] = fallingVal  #risingVal, re-use fallingVal to make line horizontal
            y[xyIdx+1] = fallingVal
            y[xyIdx+2] = np.nan

            # each spike has 3x pnts: rise/fall/nan
            xyIdx += numPerSpike  # accounts for rising/falling/nan
        # end for width
    # end for spike

    '''
    print(f't:{len(t)} {type(t)} {t.shape}')
    print(f'v:{len(v)} {type(v)} {v.shape}')
    print(f'x:{len(x)} {type(x)} {x[0:10]}')
    print(f'y:{len(y)} {type(y)} {y[0:10]}')
    '''
    #
    return x, y



from typing import Union, Dict, List, Tuple, Optional, Optional

from sanpy.sanpyLogger import get_logger
logger = get_logger(__name__)

import sanpy
from sanpy.interface.plugins import sanpyPlugin
#from sanpy.interface import kymographWidget

class kymographPlugin(sanpyPlugin):
    myHumanName = 'Kymograph'

    #def __init__(self, myAnalysisDir=None, **kwargs):
    def __init__(self, plotRawAxis=False, ba=None, **kwargs):
        """
        Args:
            ba (bAnalysis): Not required
        """
        
        logger.info('')
        
        super().__init__(ba=ba, **kwargs)

        # don't create until we get a good ba
        # if ba is None:
        #     self._kymWidget = None
        # else:
        #     self._kymWidget = sanpy.interface.kymographPlugin2(ba)
        self._kymWidget = sanpy.interface.kymographPlugin2(ba)

        self.getVBoxLayout().addWidget(self._kymWidget)
        
    # def getWidget(self):
    #     return self._kymWidget

    def replot(self):
        if self._kymWidget is None:
            self._kymWidget = sanpy.interface.kymographPlugin2(self.ba)
        self._kymWidget.slotSwitchFile(self.ba)

    def slot_switchFile(self,
                        ba : sanpy.bAnalysis,
                        rowDict : Optional[dict] = None,
                        replot : bool = True):
        # don't replot until we set our detectionClass
        replot = False
        super().slot_switchFile(ba, rowDict, replot=replot)

        if ba is not None and ba.fileLoader.isKymograph():
            logger.error(f'only tif files are supported')
            #self._initError = True
            return

        self.replot()

    # def getWidget(self):
    #     """Over-ride if plugin makes its own widget.
    #     """
    #     return self._kymWidget





"""
Plugin to create and save Axon ATF files

File looks like this 20211119

ATF    1.0
8    11
"AcquisitionMode=Episodic Stimulation"
"Comment="
"YTop=20000"
"YBottom=-20000"
"SyncTimeUnits=100"
"SweepStartTimesMS=21.000,221.000,421.000,621.000,821.000,1021.000,1221.000,1421.000,1621.000,1821.000"
"SignalsExported=IN 0"
"Signals="    "IN 0"    "IN 0"    "IN 0"    "IN 0"    "IN 0"    "IN 0"    "IN 0"    "IN 0"    "IN 0"    "IN 0"
"Time (s)"    "Trace #1 (pA)"    "Trace #2 (pA)"    "Trace #3 (pA)"    "Trace #4 (pA)"    "Trace #5 (pA)"    "Trace #6 (pA)"    "Trace #7 (pA)"    "Trace #8 (pA)"    "Trace #9 (pA)"    "Trace #10 (pA)"
0    1.2207    1.2207    5.49316    2.44141    3.66211    0.610352    2.44141    5.49316    1.83105    2.44141
1e-4    2.44141    1.2207    0.610352    2.44141    2.44141    0.610352    0.610352    1.2207    3.05176    3.05176
"""

import os, sys
from functools import partial

from datetime import datetime

import numpy as np
import pandas as pd

from PyQt5 import QtCore, QtWidgets, QtGui

import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.backends import backend_qt5agg

#import qdarkstyle

import sanpy
from sanpy.interface.plugins import sanpyPlugin

from sanpy.sanpyLogger import get_logger
logger = get_logger(__name__)

def readFolderParams(folderPath):
    """
    Return a DataFrame, one row per file. Columns are stim parameters
    """
    dList = []
    for file in sorted(os.listdir(folderPath)):
        if not file.endswith('.atf'):
            continue
        filePath = os.path.join(folderPath, file)
        d = readFileParams(filePath)
        dList.append(d)
    #
    df = pd.DataFrame(dList)
    outFile = os.path.join(folderPath, 'stimGen_db.csv')
    print('saving:', outFile)
    df.to_csv(outFile)

    return df

def readFileParams(path):
    """
    Read stimGen params from one atf file
    """
    retDict = {}
    with open(path, 'r') as f:
        while True:
            #lines = f.readlines()
            line = f.readline()
            if not line:
                break  # EOF
            if line.startswith('"Comment='):
                # looks like
                # "Comment=version=0.2;numSweeps=5;sweepDurSeconds=30.0;stimType=Sin;stimStartSeconds=5.0;durSeconds=20.0;yStimOffset=0.0;amplitude=0.002;frequency=1.0;noiseAmplitude=0.0;amplitudeStep=0.0;frequencyStep=0.0;noiseStep=0.0;doRectify=False;"
                line = line.replace('"Comment=','')
                line = line[0:-2]  # remove trailing "
                #print(line)
                for param in line.split(';'):
                    kv = param.split('=')
                    #print(kv)
                    if len(kv) != 2:
                        continue
                    k = kv[0]
                    v = kv[1]
                    #print(k, v)

                    if k == 'stimType':
                        pass
                    elif k == 'doRectify':
                        v = bool(v)
                    elif k == 'numSweeps':
                        v = int(v)
                    else:
                        v = float(v)
                    #
                    retDict[k] = v
                #
                break
    #
    return retDict

class stimGen(sanpyPlugin):
    myHumanName = 'Stim Gen'

    def __init__(self, myAnalysisDir=None, **kwargs):
        """
        Args:
            ba (bAnalysis): Not required
        """
        super(stimGen, self).__init__(**kwargs)

        self.saveStimIndex = 0

        self.version = 0.2  # v0.2 on 20211202

        self._fs = 10000
        # sampling frequency in (samples per second)

        self.numSweeps = 5

        self._data = [None] * self.numSweeps  # list of sweeps
        self._t = []

        self.stimTypes = [
                        'Sin',
                        'Chirp',
                        'Noise',
                        'Epsp train',
                        #'Stochastic HH',
                        'Integrate and Fire'
                        ]
        # TODO: add (save index, ...)
        self.stimType = 'Sin'
        self.sweepDurSeconds = 30.0
        self.stimStartSeconds = 5.0
        self.durSeconds = 20.0  # of stim
        self.yStimOffset = 0.0
        self.amplitude = 0.002
        self.frequency = 1.0
        self.noiseAmplitude = 0.0  # sqrt() of this is STD of Gaussian noise

        # step for multiple sweeps
        self.amplitudeStep = 0.0
        self.frequencyStep = 0.0
        self.noiseStep = 0.0

        self.doRectify = False
        """ remove all <0 from output stim """

        self.scale = 1
        """ scale output stim by this factor"""

        self.savePath = ''  # remember last save folder

        # TODO: all params need to be in dictionary
        # when user clicks on interface (spinbox), we need to update this dict
        self.paramDict = self.defaultParams()

        self.buildUI()

        self.updateStim()

    def defaultParams(self):
        paramDict = {
            'version': 0.3,
            'stimType': 'Sin',  # str of stim types
            'fs': 10000,  # int, samples per second
            'numSweeps': 5,  # int
            'sweepDur_sec': 30.0,
            'stimStart_sec': 5.0,
            'stimDur_sec': 20.0,
            'yOffset': 0.0,
            # First sweep parameters
            'stimFreq': 1.0,
            'stimAmp': 0.002,
            'stimNoiseAmp': 0.0,
            # step parameters
            'stimFreqStep': 0.0,
            'stimAmpStep': 0.0,
            'stimNoiseStep': 0.0,
            #
            'rectify': False,
            'scale': 1,


        }

    def getComment(self):
        comment = '' # '"Comment='

        comment +=  f'version={self.version};'
        comment +=  f'numSweeps={self.numSweeps};'
        comment +=  f'sweepDurSeconds={self.sweepDurSeconds};'
        comment +=  f'stimType={self.stimType};'
        comment +=  f'stimStartSeconds={self.stimStartSeconds};'
        comment +=  f'durSeconds={self.durSeconds};'
        comment +=  f'yStimOffset={self.yStimOffset};'
        comment +=  f'amplitude={self.amplitude};'
        comment +=  f'frequency={self.frequency};'
        comment +=  f'noiseAmplitude={self.noiseAmplitude};'
        comment +=  f'amplitudeStep={self.amplitudeStep};'
        comment +=  f'frequencyStep={self.frequencyStep};'
        comment +=  f'noiseStep={self.noiseStep};'
        comment +=  f'doRectify={self.doRectify};'

        #
        #comment += '"'
        return comment

    def getAtfHeader(self, numChannels=1):
        """
        See: https://github.com/christianrickert/Axon-Text-File/blob/master/data.atf

        File looks like this 20211119

        ATF    1.0
        8    11
        "AcquisitionMode=Episodic Stimulation"
        "Comment="
        "YTop=20000"
        "YBottom=-20000"
        "SyncTimeUnits=100"
        "SweepStartTimesMS=21.000,221.000,421.000,621.000,821.000,1021.000,1221.000,1421.000,1621.000,1821.000"
        "SignalsExported=IN 0"
        "Signals="    "IN 0"    "IN 0"    "IN 0"    "IN 0"    "IN 0"    "IN 0"    "IN 0"    "IN 0"    "IN 0"    "IN 0"
        "Time (s)"    "Trace #1 (pA)"    "Trace #2 (pA)"    "Trace #3 (pA)"    "Trace #4 (pA)"    "Trace #5 (pA)"    "Trace #6 (pA)"    "Trace #7 (pA)"    "Trace #8 (pA)"    "Trace #9 (pA)"    "Trace #10 (pA)"
        0    1.2207    1.2207    5.49316    2.44141    3.66211    0.610352    2.44141    5.49316    1.83105    2.44141
        1e-4    2.44141    1.2207    0.610352    2.44141    2.44141    0.610352    0.610352    1.2207    3.05176    3.05176
        """

        myUnits = 'pA'
        #myComment = 'fill in with stim params'

        numDataColumns = numChannels + 1  # time + number of channels
        eol = '\n'
        tab = '\t'
        ATF_HEADER = "ATF    1.0" + eol
        ATF_HEADER += f'8\t{numDataColumns}' + eol

        ATF_HEADER += '"AcquisitionMode=Episodic Stimulation"' + eol

        myComment = self.getComment()
        ATF_HEADER += f'"Comment={myComment}"' + eol

        ATF_HEADER += '"YTop=20000"' + eol
        ATF_HEADER += '"YBottom=-20000"' + eol
        ATF_HEADER += '"SyncTimeUnits=100"' + eol

        # for a 200 ms sweep, looks like this
        # "SweepStartTimesMS=21.000,221.000,421.000,621.000,821.000,1021.000,1221.000,1421.000,1621.000,1821.000"
        #durMs = self.durSeconds *1000
        durMs = self.sweepDurSeconds *1000  # new 20211202
        numSweeps = self.numSweeps
        sweepRange = range(numSweeps)
        SweepStartTimesMS = '"SweepStartTimesMS='
        preRollMs = 21
        for sweepIdx, sweep in enumerate(sweepRange):
            currStartTime = preRollMs + (sweepIdx * durMs)
            SweepStartTimesMS += f'{round(float(currStartTime),3)}' + ','
        # remove last comma
        SweepStartTimesMS = SweepStartTimesMS[0:-1]
        SweepStartTimesMS += '"'
        ATF_HEADER += SweepStartTimesMS + eol

        ATF_HEADER += '"SignalsExported=IN 0"' + eol

        # signals looks like this
        # "Signals="    "IN 0"    "IN 0"    "IN 0"    "IN 0"    "IN 0"    "IN 0"    "IN 0"    "IN 0"    "IN 0"    "IN 0"
        Signals = '"Signals="' + '\t'
        for sweepIdx, sweep in enumerate(sweepRange):
            Signals += '"IN 0"' + '\t'
        # remove last tab
        Signals = Signals[0:-1]
        ATF_HEADER += Signals + eol

        ATF_HEADER += f'"Time (s)"' + tab
        traceStr = ''
        for channel in range(numChannels):
            traceStr += f'"Trace #{channel+1} ({myUnits})"' + tab  # not sure if trailing tab is ok???
        # remove last tab
        traceStr = traceStr[0:-1]
        ATF_HEADER += traceStr + eol

        print('=== ATF_HEADER:')
        print(ATF_HEADER)

        '''
        ATF_HEADER = """
        ATF    1.0
        8    3
        "AcquisitionMode=Episodic Stimulation"
        "Comment="
        "YTop=2000"
        "YBottom=-2000"
        "SyncTimeUnits=20"
        "SweepStartTimesMS=0.000"
        "SignalsExported=IN 0"
        "Signals="    "IN 0"
        "Time (s)"    "Trace #1"    "Trace #2"
        """.strip()
        '''

        return ATF_HEADER

    '''
    @property
    def data(self):
        return self._data
    '''

    @property
    def t(self):
        return self._t

    @property
    def fs(self):
        return self._fs

    def makeStim(self):
        type = self.stimType
        fs = self.fs
        sweepDurSeconds = self.sweepDurSeconds  # new 20211202
        stimStartSeconds = self.stimStartSeconds  # of stim
        durSec = self.durSeconds  # of stim
        yStimOffset = self.yStimOffset
        amp = self.amplitude
        freq = self.frequency
        noiseAmp = self.noiseAmplitude
        doRectify = self.doRectify

        amplitudeStep = self.amplitudeStep
        frequencyStep = self.frequencyStep
        noiseStep = self.noiseStep

        self._data = [None] * self.numSweeps
        for sweepNum in range(self.numSweeps):
            currAmp = amp + (sweepNum * amplitudeStep)
            currFreq = freq + (sweepNum * frequencyStep)
            currNoiseAmp = noiseAmp + (sweepNum * noiseStep)
            autoPad = False
            print(f'  makeStim() {type} sweep:{sweepNum} durSec:{durSec} amp:{currAmp} freq:{currFreq} noiseAmp:{currNoiseAmp}')
            self._data[sweepNum] = sanpy.atfStim.makeStim(type, sweepDurSec=sweepDurSeconds,
                            startStimSec=stimStartSeconds, durSec=durSec,
                            yStimOffset=yStimOffset, amp=currAmp,
                            freq=currFreq, fs=fs, noiseAmp=currNoiseAmp, rectify=doRectify,
                            autoPad=autoPad, autoSave=False)
            if self._data[sweepNum] is None:
                print(f'makeStim() error making {type} at sweep number {sweepNum}')
            #
            # scale
            self._data[sweepNum] *= self.scale
            print('  self._data[sweepNum].dtype:', self._data[sweepNum].dtype)

        self._t = np.arange(len(self._data[0])) / fs  # just using first sweep

    '''
    def plotStim(self):
        logger.info(f'_t:{self._t.shape}')
        logger.info(f'_data:{self._data.shape}')
        plt.plot(self._t, self._data)
    '''

    '''
    def saveStim(Self):
        sanpy.atfStim.saveAtf(self.data, fileName="output.atf", fs=10000)
    '''

    def _grabParams(self):
        """
        Grab all interface parameters in member variables
        """

        self.sweepDurSeconds = self.sweepDurationSpinBox.value()  # new 20211202
        self.numSweeps = self.numSweepsSpinBox.value()

        self.stimStartSeconds = self.stimStartSpinBox.value()  # start of sin
        self.durSeconds = self.durationSpinBox.value()  # duration of stim
        self.yStimOffset = self.yStimOffsetSpinBox.value()
        self.amplitude = self.amplitudeSpinBox.value()
        self.frequency = self.frequencySpinBox.value()
        self.noiseAmplitude = self.noiseAmpSpinBox.value()

        self.amplitudeStep = self.amplitudeStepSpinBox.value()
        self.frequencyStep = self.frequencyStepSpinBox.value()
        self.noiseStep = self.noiseStepSpinBox.value()

        self.doRectify = self.rectifyCheckBox.isChecked()
        self._fs = self.fsSpinBox.value()

        '''
        rmsMult = 1/np.sqrt(2)
        sinRms = self.amplitude * rmsMult
        sinRms = round(sinRms,2)
        aName = f'RMS:{sinRms}'
        self.sinRms.setText(aName)
        '''

    def updateStim(self):
        self._grabParams()
        self.makeStim()
        self.replot()

    def on_spin_box2(self, name, obj):
        print(name, obj.value())

    def on_spin_box(self, name):
        logger.info(name)
        if name == 'Number Of Sweeps':
            numSweeps = self.numSweepsSpinBox.value()
            self._updateNumSweeps(numSweeps)
        elif name == 'Save Index':
            saveStimIndex = self.saveIndexSpinBox.value()
            self.saveStimIndex = saveStimIndex
            #
        self.updateStim()

    def on_stim_type(self, type):
        logger.info(type)
        self.stimType = type
        self.updateStim()

    def on_scale(self, scale):
        logger.info(type)
        self.scale = float(scale)
        self.updateStim()

    def on_button_click(self, name):
        logger.info(name)
        if name == 'Make Stimulus':
            self.makeStim()
        elif name == 'Save As...':
            # TODO: srt a feeback red as we save, set to green when done
            # self.saveAsButton
            self.saveAs()
        else:
            logger.info(f'name "{name}" not understood.')

    def on_checkbox_clicked(self, name):
        self.updateStim()

    def buildUI(self):
        # main layout
        vLayout = QtWidgets.QVBoxLayout()
        controlLayout = QtWidgets.QHBoxLayout()

        '''
        aName = 'Make Stimulus'
        aButton = QtWidgets.QPushButton(aName)
        aButton.clicked.connect(partial(self.on_button_click,aName))
        controlLayout.addWidget(aButton)
        '''

        aName = 'Save As...'
        self.saveAsButton = QtWidgets.QPushButton(aName)
        self.saveAsButton.clicked.connect(partial(self.on_button_click,aName))
        controlLayout.addWidget(self.saveAsButton)

        aName = 'Save Index'
        aLabel = QtWidgets.QLabel(aName)
        controlLayout.addWidget(aLabel)
        self.saveIndexSpinBox = QtWidgets.QSpinBox()
        self.saveIndexSpinBox.setKeyboardTracking(False)
        self.saveIndexSpinBox.setRange(0, 9999)
        self.saveIndexSpinBox.setValue(self.saveStimIndex)
        self.saveIndexSpinBox.valueChanged.connect(partial(self.on_spin_box, aName))
        #self.saveIndexSpinBox.valueChanged.connect(partial(self.on_spin_box2, aName, self.saveIndexSpinBox))
        controlLayout.addWidget(self.saveIndexSpinBox)

        aName = 'Stim Type'
        aLabel = QtWidgets.QLabel(aName)
        controlLayout.addWidget(aLabel)
        self.stimTypeDropdown = QtWidgets.QComboBox()
        for type in self.stimTypes:
            self.stimTypeDropdown.addItem(type)
        self.stimTypeDropdown.currentTextChanged.connect(self.on_stim_type)
        controlLayout.addWidget(self.stimTypeDropdown)

        aName = 'Sweep Duration(s)'
        aLabel = QtWidgets.QLabel(aName)
        controlLayout.addWidget(aLabel)
        self.sweepDurationSpinBox = QtWidgets.QDoubleSpinBox()
        self.sweepDurationSpinBox.setKeyboardTracking(False)
        self.sweepDurationSpinBox.setRange(0, 2**16)
        self.sweepDurationSpinBox.setValue(self.sweepDurSeconds)
        self.sweepDurationSpinBox.valueChanged.connect(partial(self.on_spin_box, aName))
        controlLayout.addWidget(self.sweepDurationSpinBox)

        aName = 'Number Of Sweeps'
        aLabel = QtWidgets.QLabel(aName)
        controlLayout.addWidget(aLabel)
        self.numSweepsSpinBox = QtWidgets.QSpinBox()
        self.numSweepsSpinBox.setKeyboardTracking(False)
        self.numSweepsSpinBox.setRange(1, 100)
        self.numSweepsSpinBox.setValue(self.numSweeps)
        self.numSweepsSpinBox.valueChanged.connect(partial(self.on_spin_box, aName))
        controlLayout.addWidget(self.numSweepsSpinBox)

        vLayout.addLayout(controlLayout) # add mpl canvas

        # row 1.5
        controlLayout_row1_5 = QtWidgets.QHBoxLayout()

        aName = 'Stim Start(s)'
        aLabel = QtWidgets.QLabel(aName)
        controlLayout_row1_5.addWidget(aLabel)
        self.stimStartSpinBox = QtWidgets.QDoubleSpinBox()
        self.stimStartSpinBox.setKeyboardTracking(False)
        self.stimStartSpinBox.setRange(0, 2**16)
        self.stimStartSpinBox.setValue(self.stimStartSeconds)
        self.stimStartSpinBox.valueChanged.connect(partial(self.on_spin_box, aName))
        controlLayout_row1_5.addWidget(self.stimStartSpinBox)

        aName = 'Stim Duration(s)'
        aLabel = QtWidgets.QLabel(aName)
        controlLayout_row1_5.addWidget(aLabel)
        self.durationSpinBox = QtWidgets.QDoubleSpinBox()
        self.durationSpinBox.setKeyboardTracking(False)
        self.durationSpinBox.setRange(0, 2**16)
        self.durationSpinBox.setValue(self.durSeconds)
        self.durationSpinBox.valueChanged.connect(partial(self.on_spin_box, aName))
        controlLayout_row1_5.addWidget(self.durationSpinBox)

        aName = 'Stim Offset(y)'
        aLabel = QtWidgets.QLabel(aName)
        controlLayout_row1_5.addWidget(aLabel)
        self.yStimOffsetSpinBox = QtWidgets.QDoubleSpinBox()
        self.yStimOffsetSpinBox.setKeyboardTracking(False)
        self.yStimOffsetSpinBox.setRange(-1e9, 2**16)
        self.yStimOffsetSpinBox.setValue(self.yStimOffset)
        self.yStimOffsetSpinBox.valueChanged.connect(partial(self.on_spin_box, aName))
        controlLayout_row1_5.addWidget(self.yStimOffsetSpinBox)

        vLayout.addLayout(controlLayout_row1_5) # add mpl canvas

        # 2nd row
        controlLayout_row2 = QtWidgets.QGridLayout()
        rowSpan = 1
        colSpan = 1
        row = 0
        col = 0

        aName = 'Amplitude'
        aLabel = QtWidgets.QLabel(aName)
        controlLayout_row2.addWidget(aLabel, row, col, rowSpan, colSpan)
        self.amplitudeSpinBox = QtWidgets.QDoubleSpinBox()
        self.amplitudeSpinBox.setKeyboardTracking(False)
        self.amplitudeSpinBox.setRange(0, 2**16)
        self.amplitudeSpinBox.setDecimals(5)
        self.amplitudeSpinBox.setValue(self.amplitude)
        self.amplitudeSpinBox.valueChanged.connect(partial(self.on_spin_box, aName))
        controlLayout_row2.addWidget(self.amplitudeSpinBox, 0, 1, rowSpan, colSpan)

        # rms of sin (amp and freq)
        '''
        rmsMult = 1/np.sqrt(2)
        sinRms = self.amplitude * rmsMult
        sinRms = round(sinRms,2)
        aName = f'RMS:{sinRms}'
        self.sinRms = QtWidgets.QLabel(aName)
        controlLayout_row2.addWidget(self.sinRms, 0, 2, rowSpan, colSpan)
        '''

        aName = 'Frequency (Hz)'
        aLabel = QtWidgets.QLabel(aName)
        controlLayout_row2.addWidget(aLabel, 0, 3, rowSpan, colSpan)
        self.frequencySpinBox = QtWidgets.QDoubleSpinBox()
        self.frequencySpinBox.setKeyboardTracking(False)
        self.frequencySpinBox.setRange(0, 2**16)
        self.frequencySpinBox.setValue(self.frequency)
        self.frequencySpinBox.valueChanged.connect(partial(self.on_spin_box, aName))
        controlLayout_row2.addWidget(self.frequencySpinBox, 0, 4, rowSpan, colSpan)

        aName = 'Noise Amplitude'
        aLabel = QtWidgets.QLabel(aName)
        controlLayout_row2.addWidget(aLabel, 0, 5, rowSpan, colSpan)
        self.noiseAmpSpinBox = QtWidgets.QDoubleSpinBox()
        self.noiseAmpSpinBox.setKeyboardTracking(False)
        self.noiseAmpSpinBox.setDecimals(5)
        self.noiseAmpSpinBox.setSingleStep(0.01)
        self.noiseAmpSpinBox.setRange(0, 2**16)
        self.noiseAmpSpinBox.setValue(self.noiseAmplitude)
        self.noiseAmpSpinBox.valueChanged.connect(partial(self.on_spin_box, aName))
        controlLayout_row2.addWidget(self.noiseAmpSpinBox, 0, 6, rowSpan, colSpan)

        #
        # row 2
        #controlLayout_row2 = QtWidgets.QHBoxLayout()

        aName = 'Amplitude Step'
        aLabel = QtWidgets.QLabel(aName)
        controlLayout_row2.addWidget(aLabel, 1, 0, rowSpan, colSpan)
        self.amplitudeStepSpinBox = QtWidgets.QDoubleSpinBox()
        self.amplitudeStepSpinBox.setKeyboardTracking(False)
        self.amplitudeStepSpinBox.setDecimals(5)
        self.amplitudeStepSpinBox.setSingleStep(0.01)
        self.amplitudeStepSpinBox.setRange(0, 2**16)
        self.amplitudeStepSpinBox.setValue(self.amplitudeStep)
        self.amplitudeStepSpinBox.valueChanged.connect(partial(self.on_spin_box, aName))
        controlLayout_row2.addWidget(self.amplitudeStepSpinBox, 1, 1, rowSpan, colSpan)

        aName = 'Frequency Step'
        aLabel = QtWidgets.QLabel(aName)
        controlLayout_row2.addWidget(aLabel, 1, 3, rowSpan, colSpan)
        self.frequencyStepSpinBox = QtWidgets.QDoubleSpinBox()
        self.frequencyStepSpinBox.setKeyboardTracking(False)
        self.frequencyStepSpinBox.setSingleStep(0.1)
        self.frequencyStepSpinBox.setRange(0, 2**16)
        self.frequencyStepSpinBox.setValue(self.frequencyStep)
        self.frequencyStepSpinBox.valueChanged.connect(partial(self.on_spin_box, aName))
        controlLayout_row2.addWidget(self.frequencyStepSpinBox, 1, 4, rowSpan, colSpan)

        # first row in grid has freq rms

        aName = 'Noise Step'
        aLabel = QtWidgets.QLabel(aName)
        controlLayout_row2.addWidget(aLabel, 1, 5, rowSpan, colSpan)
        self.noiseStepSpinBox = QtWidgets.QDoubleSpinBox()
        self.noiseStepSpinBox.setKeyboardTracking(False)
        self.noiseStepSpinBox.setDecimals(5)
        self.noiseStepSpinBox.setSingleStep(0.01)
        self.noiseStepSpinBox.setRange(0, 2**16)
        self.noiseStepSpinBox.setValue(self.frequencyStep)
        self.noiseStepSpinBox.valueChanged.connect(partial(self.on_spin_box, aName))
        controlLayout_row2.addWidget(self.noiseStepSpinBox, 1, 6, rowSpan, colSpan)

        #
        vLayout.addLayout(controlLayout_row2) # add mpl canvas

        controlLayout_row3 = QtWidgets.QHBoxLayout()

        checkboxName = 'Rectify'
        self.rectifyCheckBox = QtWidgets.QCheckBox(checkboxName)
        self.rectifyCheckBox.setChecked(self.doRectify)
        self.rectifyCheckBox.stateChanged.connect(partial(self.on_checkbox_clicked, checkboxName))
        controlLayout_row3.addWidget(self.rectifyCheckBox)

        aName = 'Samples Per Second'
        aLabel = QtWidgets.QLabel(aName)
        controlLayout_row3.addWidget(aLabel)
        self.fsSpinBox = QtWidgets.QSpinBox()
        self.fsSpinBox.setKeyboardTracking(False)
        self.fsSpinBox.setRange(1, 100000)  # TODO: Fix hard coding of 100000
        self.fsSpinBox.setValue(self._fs)
        self.fsSpinBox.valueChanged.connect(partial(self.on_spin_box, aName))
        controlLayout_row3.addWidget(self.fsSpinBox)

        aName = 'Scale'
        aLabel = QtWidgets.QLabel(aName)
        controlLayout_row3.addWidget(aLabel)
        self.scaleDropdown = QtWidgets.QComboBox()
        scales = [0.001, 0.01, 0.1, 1, 10, 100, 1000]
        for scale in scales:
            scaleStr = str(scale)
            self.scaleDropdown.addItem(scaleStr)
        startIndex = 3
        self.scaleDropdown.setCurrentIndex(startIndex)
        self.scaleDropdown.currentTextChanged.connect(self.on_scale)
        controlLayout_row3.addWidget(self.scaleDropdown)

        #
        vLayout.addLayout(controlLayout_row3) # add mpl canvas

        vSplitter = QtWidgets.QSplitter(QtCore.Qt.Vertical)
        vLayout.addWidget(vSplitter)

        # plt.style.use('dark_background')
        if self.darkTheme:
            plt.style.use('dark_background')
        else:
            plt.rcParams.update(plt.rcParamsDefault)

        self.fig = mpl.figure.Figure(constrained_layout=True)
        self.static_canvas = backend_qt5agg.FigureCanvas(self.fig)
        self.static_canvas.setFocusPolicy( QtCore.Qt.ClickFocus )
        self.static_canvas.setFocus()

        #can do self.mplToolbar.hide()
        self.mplToolbar = mpl.backends.backend_qt5agg.NavigationToolbar2QT(self.static_canvas, self.static_canvas)

        self._updateNumSweeps(self.numSweeps)
        #self.rawAxes = self.static_canvas.figure.add_subplot(self.numSweeps,1,1)
        #self.plotLine, = self.rawAxes[0].plot([], [], '-w', linewidth=1)

        vSplitter.addWidget(self.static_canvas) # add mpl canvas
        vSplitter.addWidget(self.mplToolbar) # add mpl canvas

        #
        # finalize
        #self.mainWidget = QtWidgets.QWidget()
        # if qdarkstyle is not None:
        #     self.setStyleSheet(qdarkstyle.load_stylesheet(qt_api='pyqt5'))
        # else:
        #     self.setStyleSheet("")

        # set the layout of the main window
        # self.setLayout(vLayout)
        self.getVBoxLayout().addLayout(vLayout)

    def _updateNumSweeps(self, numSweeps):
        """
        Remove and add all Figure axes
        """
        self.numSweeps = numSweeps

        self.static_canvas.figure.clear()

        self.plotLine = [None] * numSweeps
        self.rawAxes = [None] * numSweeps

        for i in range(numSweeps):
            if i == 0:
                self.rawAxes[i] = self.static_canvas.figure.add_subplot(numSweeps,1, i+1)  # +1 because subplot index is 1 based
            else:
                self.rawAxes[i] = self.static_canvas.figure.add_subplot(numSweeps,1, i+1, sharex=self.rawAxes[0])  # +1 because subplot index is 1 based
            self.plotLine[i], = self.rawAxes[i].plot([], [], '-w', linewidth=0.5)

            self.rawAxes[i].spines['right'].set_visible(False)
            self.rawAxes[i].spines['top'].set_visible(False)

            lastSweep = i == (numSweeps - 1)
            if not lastSweep:
                self.rawAxes[i].spines['bottom'].set_visible(False)
                self.rawAxes[i].tick_params(axis="x", labelbottom=False) # no labels

    def replot(self):
        logger.info(f't:{len(self._t)}, data:{len(self._data)}')

        yMin = 1e9
        yMax = -1e9

        for i in range(self.numSweeps):
            self.plotLine[i].set_xdata(self._t)
            self.plotLine[i].set_ydata(self._data[i])
            #
            self.rawAxes[i].relim()
            self.rawAxes[i].autoscale_view(True,True,True)

            thisMin = np.nanmin(self._data[i])
            thisMax = np.nanmax(self._data[i])
            if thisMin < yMin:
                yMin = thisMin
            if thisMax > yMax:
                yMax = thisMax

        for i in range(self.numSweeps):
            self.rawAxes[i].set_ylim([yMin, yMax])

        #
        self.static_canvas.draw()

    '''
    def _getTime(self):
        """Get time in seconds."""
        n = int(durSec * fs) # total number of samples
        t = np.linspace(0, self.durSeconds, n, endpoint=True)
        return t
    '''

    def saveAs(self):
        """Save a stimulus waveform array as an ATF 1.0 file.

        If use specifies .atf then save as .atf
        If user specifies .csv then save as .csv
        """
        fileName = self.getFileName()
        options = QtWidgets.QFileDialog.Options()
        savePath = os.path.join(self.savePath, fileName)
        fileName, _ = QtWidgets.QFileDialog.getSaveFileName(self,"Save .atf file",
                            savePath,"Atf Files (*.atf);;CSV Files (*.csv)", options=options)
        if not fileName:
            return

        self.savePath = os.path.split(fileName)[0]

        if fileName.endswith('.atf'):
            numSweeps = self.numSweeps
            out = self.getAtfHeader(numChannels=numSweeps)
            data = self._data  # list of sweeps
            fs = self._fs

            #if numChannels == 2:
            #    myNoise = np.random.normal(scale=np.sqrt(5), size=data.shape)

            # TODO: getAtfHeader needs to append trailing eol
            #         Then here we don't pre-pend with \n but append each line
            eol = '\n'
            pntsPerSweep = len(self._data[0])
            for i in range(pntsPerSweep):
                for sweepNumber in range(self.numSweeps):
                    #sweepData = self._data[sweepNumber]
                    val = self._data[sweepNumber][i]
                    # TODO: Convert to f'' format
                    if sweepNumber == 0:
                        # time and sweep 0
                        out += '%.05f\t%.05f'%(i/fs,val)
                    else:
                        # append value for next sweep
                        out += '\t%.05f'%(val)
                #
                out += eol
            #
            with open(fileName,'w') as f:
                f.write(out)

        elif fileName.endswith('.csv'):
            df = pd.DataFrame(columns=['sec', 'pA'])
            df['sec'] = self._t
            df['pA'] = self._data
            df.to_csv(fileName, index=False)
        #
        logger.info(f'Saved: "{fileName}"')

    def getFileName(self):
        stimType = self.stimType
        numSweeps = self.numSweeps
        durSeconds = self.durSeconds  # sweep duration
        amplitude = self.amplitude
        frequency = self.frequency
        noiseAmplitude = self.noiseAmplitude
        noiseStep = self.noiseStep
        """
        _s : number of sweeps
        _sd : sweep duration (seconds)
        _a : amplitude
        _f : frequency
        _g : start noise amplitude
        _ns : noise step
        """

        """
        filename = f'{stimType}_s{numSweeps}_sd{durSeconds}_a{amplitude}_f{frequency}_g{noiseAmplitude}'
        if numSweeps > 1:
            filename += f'_ns{noiseStep}'
        filename += '.atf'
        return filename
        """

        filename = 'sanpy_' + datetime.today().strftime('%Y%m%d')
        filename += '_'

        saveStimIndex = self.saveStimIndex
        filename += f'{saveStimIndex:04}'

        filename += '.atf'

        # increment for next save
        self.saveStimIndex += 1

        # update interface
        self.saveIndexSpinBox.setValue(self.saveStimIndex)

        return filename

    def buildFromDict(self, d):
        """
        build from dict constructed from saved abt file using readParams()

        TODO: Does not work because we do not update Qt interface with values from dict
        """
        myVars = vars(self)
        print(myVars)
        for k,v in d.items():
            print(k, v, type(v))
            myVars[k] = v

        self.replot()

def run():
    app = QtWidgets.QApplication(sys.argv)

    sg = stimGen()
    sg.show()

    sys.exit(app.exec_())

def testDict():
    folderPath = '/home/cudmore/Sites/SanPy'
    df = readFolderParams(folderPath)

    sys.exit(1)

    path = '/home/cudmore/Sites/SanPy/sanpy_20211206_0000.atf'
    path = '/home/cudmore/Sites/SanPy/sanpy_20211206_0001.atf'

    d = readFileParams(path)
    for k,v in d.items():
        print(k,v)

    '''
    app = QtWidgets.QApplication(sys.argv)
    sg = stimGen()
    sg.buildFromDict(d)

    sg.show()

    sys.exit(app.exec_())
    '''

if __name__ == '__main__':
    run()

    #testDict()
